from datetime import date

from nose.tools import eq_

from inferno.lib.job_options import JobOptions
from inferno.lib.rule import InfernoRule
from inferno.lib.settings import InfernoSettings


class TestJobOptions(object):

    def setUp(self):
        self.rule = InfernoRule(
            day_range=3,
            day_offset=1,
            day_start=date(2012, 12, 02),
            source_tags=['tag1', 'tag2'],
            result_tag='result_tag_rule')
        self.settings = InfernoSettings(
            day_range=4,
            day_offset=2,
            day_start=date(2011, 12, 02),
            source_tags=['tag3', 'tag4'],
            result_tag='result_tag_settings')

        # expected results
        self.result_tag_from_rule = 'result_tag_rule'
        self.result_tag_from_settings = 'result_tag_settings'
        self.tags_from_rule = [
                'tag1:2012-12-01',
                'tag1:2012-11-30',
                'tag1:2012-11-29',
                'tag2:2012-12-01',
                'tag2:2012-11-30',
                'tag2:2012-11-29']
        self.tags_from_settings = [
                'tag3:2011-11-30',
                'tag3:2011-11-29',
                'tag3:2011-11-28',
                'tag3:2011-11-27',
                'tag4:2011-11-30',
                'tag4:2011-11-29',
                'tag4:2011-11-28',
                'tag4:2011-11-27']

    def test_empty_rule_and_empty_settings(self):
        job_options = JobOptions(InfernoRule(), InfernoSettings())
        eq_(job_options.tags, [])
        eq_(job_options.result_tag, None)

    def test_result_tag_from_rule(self):
        actual = JobOptions(self.rule, InfernoSettings()).result_tag
        eq_(actual, self.result_tag_from_rule)

    def test_result_tag_from_settings(self):
        actual = JobOptions(InfernoRule(), self.settings).result_tag
        eq_(actual, self.result_tag_from_settings)

    def test_result_tag_settings_trump_rule(self):
        actual = JobOptions(self.rule, self.settings).result_tag
        eq_(actual, self.result_tag_from_settings)

    def test_tags_from_rule(self):
        actual = JobOptions(self.rule, InfernoSettings()).tags
        eq_(actual, self.tags_from_rule)

    def test_tags_from_settings(self):
        actual = JobOptions(InfernoRule(), self.settings).tags
        eq_(actual, self.tags_from_settings)

    def test_tags_settings_trump_rule(self):
        actual = JobOptions(self.rule, self.settings).tags
        eq_(actual, self.tags_from_settings)

    def test_tags_from_settings_and_rule_mix(self):
        rule = InfernoRule(source_tags=['tag5'], day_range=2)
        settings = InfernoSettings(day_start=date(2011, 12, 01))
        actual = JobOptions(rule, settings).tags
        expected = ['tag5:2011-12-01', 'tag5:2011-11-30']
        eq_(actual, expected)

    def test_explict_tags(self):
        rule = InfernoRule(source_tags=['tag:foo', 'tag:bar'])
        settings = InfernoSettings()
        actual = JobOptions(rule, settings).tags
        expected = ['tag:foo', 'tag:bar']
        eq_(actual, expected)

    def test_explict_tags_despite_day_range_on_the_rule(self):
        rule = InfernoRule(source_tags=['tag:foo', 'tag:bar'], day_range=2)
        settings = InfernoSettings(day_range=0)
        actual = JobOptions(rule, settings).tags
        expected = ['tag:foo', 'tag:bar']
        eq_(actual, expected)



from nose.tools import eq_

from disco.core import result_iterator

from inferno.lib.rule import InfernoRule
from inferno.lib.rule import Keyset
from inferno.lib.disco_ext import sorted_iterator


class TestInfernoRule(object):

    def test_keysets(self):
        # no key sets
        rule = InfernoRule()
        eq_(rule.params.keysets, {})

        # one key set
        rule = InfernoRule(
            key_parts=['id'],
            value_parts=['count'],
            table='some_table',
            column_mappings={'id': 'some_id'})
        keysets = {
            '_default': {
                'column_mappings': {'id': 'some_id'},
                'table': 'some_table',
                'value_parts': ['count'],
                'key_parts': ['_keyset', 'id']}}
        eq_(rule.params.keysets, keysets)

        # many key sets
        rule = InfernoRule(
            keysets={
                'keyset1': Keyset(
                    key_parts=['id1'],
                    value_parts=['count1'],
                    column_mappings={'id1': 'some_id1'},
                    table='some_table1'),
                'keyset2': Keyset(
                    key_parts=['id2'],
                    value_parts=['count2'],
                    column_mappings={'id2': 'some_id2'},
                    table='some_table2')})
        keysets = {
            'keyset1': {
                'column_mappings': {'id1': 'some_id1'},
                'table': 'some_table1',
                'value_parts': ['count1'],
                'key_parts': ['_keyset', 'id1']},
            'keyset2': {
                'column_mappings': {'id2': 'some_id2'},
                'table': 'some_table2',
                'value_parts': ['count2'],
                'key_parts': ['_keyset', 'id2']}}
        eq_(rule.params.keysets, keysets)

    def test_parts_preprocess(self):
        def foo(parts, params):
            parts['bar'] = 1
            yield parts

        rule = InfernoRule(parts_preprocess=[foo])
        eq_(rule.params.parts_preprocess, ['foo'])
        actual = rule.params.foo({'hello': 'world'}, None)
        eq_(list(actual), [{'bar':1, 'hello':'world'}])

    def test_field_transforms(self):
        def upper(val):
            return val.upper()

        rule = InfernoRule(field_transforms={'hello': upper})
        eq_(rule.params.field_transforms, {'hello': 'upper'})
        eq_(rule.params.upper('world'), 'WORLD')

    def test_result_iterator(self):
        # sort=default
        rule = InfernoRule()
        eq_(rule.result_iterator, sorted_iterator)

        # sort=True
        rule = InfernoRule(sort=True)
        eq_(rule.result_iterator, sorted_iterator)

        # sort=False
        rule = InfernoRule(sort=False)
        eq_(rule.result_iterator, result_iterator)

    def test_source_tags(self):
        # list
        rule = InfernoRule(source_tags=['tag1', 'tag2'])
        eq_(rule.source_tags, ['tag1', 'tag2'])

        # empty list
        rule = InfernoRule(source_tags=[])
        eq_(rule.source_tags, [])

        # one tag (string)
        rule = InfernoRule(source_tags='tag1')
        eq_(rule.source_tags, ['tag1'])

        # none tag
        rule = InfernoRule(source_tags=None)
        eq_(rule.source_tags, None)

    def test_kwargs(self):
        rule = InfernoRule(some_extra_param='some_extra_value')
        eq_(rule.params.some_extra_param, 'some_extra_value')

    def test_str(self):
        rule = InfernoRule(name='some_rule_name')
        eq_(str(rule), '<InfernoRule: some_rule_name>')



import collections


class peekable(object):
    """
    THIS CODE IS FROM THE PYTHON COOKBOOK.
    """

    def __init__(self, iterable):
        self._iterable = iter(iterable)
        self._cache = collections.deque()

    def __iter__(self):
        return self

    def _fillcache(self, n):
        if n is None:
            n = 1
        while len(self._cache) < n:
            self._cache.append(self._iterable.next())

    def next(self, n=None):
        self._fillcache(n)
        if n is None:
            result = self._cache.popleft()
        else:
            result = [self._cache.popleft() for i in range(n)]
        return result

    def peek(self, n=None):
        self._fillcache(n)
        if n is None:
            result = self._cache[0]
        else:
            result = [self._cache[i] for i in range(n)]
        return result



def keyset_map(parts_, params_):

    import ujson

    class KeysetMap(object):

        def __init__(self, params):
            self.params = params

        def __call__(self, parts):
            try:
                self._debug('input: %s', parts)
                for parts in self._preprocess([parts]):
                    self._transform(parts)
                    result = self._result(parts)
                    if result is not None:
                        self._debug('result: %s', result)
                        yield result
            except Exception:
                self._error('input: %s', parts_)

        def _preprocess(self, parts_list):
            if hasattr(self.params, 'parts_preprocess'):
                # each preprocessor may generate multiple 'parts',
                # these need to be fed into subsequent preprocessors
                for name in self.params.parts_preprocess:
                    func = getattr(self.params, name)
                    new_list = []
                    for parts in parts_list:
                        new_list.extend([x for x in func(parts, self.params)])
                    parts_list = new_list
            return parts_list

        def _transform(self, parts):
            if hasattr(self.params, 'field_transforms'):
                for field, name in self.params.field_transforms.items():
                    func = getattr(self.params, name)
                    if field in parts:
                        parts[field] = func(parts[field])

        def _result(self, parts):
            keyset = parts.get('_keyset', '_default')
            if keyset in self.params.keysets:
                key_parts = self.params.keysets[keyset]['key_parts']
                value_parts = self.params.keysets[keyset]['value_parts']
                key = tuple(self._make_key(parts, key_parts))
                if None not in key[1:]:
                    value = [parts.get(a, 0) for a in value_parts]
                    try:
                        return ujson.dumps(key), value
                    except Exception:
                        self._error(': %s', parts_)

        def _make_key(self, parts, keys):
            result = []
            for key in keys:
                if key in parts and parts[key] is not None:
                    result.append(self._safe_str(parts[key]))
                else:
                    result.append(None)
            return result

        def _safe_str(self, value):
            try:
                return str(value)
            except UnicodeEncodeError:
                return unicode(value).encode('utf-8')

        def _debug(self, message, *args):
            if getattr(self.params, 'disco_debug', False):
                self._disco_message(message % args)

        def _error(self, message, *args):
            import traceback
            trace = traceback.format_exc(15)
            self._disco_message('%s %s' % (message, trace))

        def _disco_message(self, message):
            import disco.util
            disco.util.msg(message)

    go = KeysetMap(params_)
    return go(parts_)



from datetime import timedelta


class JobOptions(object):

    def __init__(self, rule, settings):
        self.rule = rule
        self.settings = settings

    @property
    def result_tag(self):
        if self.settings.get('result_tag'):
            return self.settings.get('result_tag')
        else:
            return self.rule.result_tag

    @property
    def tags(self):
        tags = []

        if self.settings.get('source_tags') is not None:
            tags = self.settings.get('source_tags')
        else:
            tags = self.rule.source_tags

        if self.settings.get('day_start') is not None:
            start = self.settings.get('day_start')
        else:
            start = self.rule.day_start

        if self.settings.get('day_offset') is not None:
            start += timedelta(days=-self.settings.get('day_offset'))
        else:
            start += timedelta(days=-self.rule.day_offset)

        if self.settings.get('day_range') is not None:
            count = self.settings.get('day_range')
        else:
            count = self.rule.day_range

        if count and tags:
            tags = [self._name(tag, day, start)
                for tag in tags for day in range(count)]

        return [] if tags is None else tags

    def _name(self, tag, delta, start):
        return '%s:%s' % (tag, start + timedelta(days=-delta))



def keyset_reduce(iter_, params_):

    import ujson

    import disco.util

    class KeysetReduce(object):

        def __init__(self, params):
            self.params = params

        def __call__(self, iter):
            for key, value in disco.util.kvgroup(iter):
                try:
                    key = ujson.loads(key)
                    sum_ = self._sum_group(key, value)
                    if (hasattr(self.params, 'serial_out') and
                        self.params.serial_out):
                        serial = ','.join([self._safe_str(y) for y in key[1:]])
                        result = serial, ujson.dumps(sum_)
                    else:
                        result = key, sum_
                    self._debug('result: %s', result)
                    yield result
                except Exception:
                    self._error('input: %s,%s', key, value)

        def _sum_group(self, key, value):
            summed_values = []
            for row in value:
                index = 0
                self._debug('input: %s,%s', key, row)
                for item in row:
                    item = self._convert_to_numeric(item)
                    if index >= len(summed_values):
                        summed_values.append(item)
                    else:
                        summed_values[index] += item
                    index += 1
            return summed_values

        def _convert_to_numeric(self, val):
            if isinstance(val, (float, int, long)):
                return val
            else:
                try:
                    return float(val)
                except:
                    return 0

        def _safe_str(self, value):
            try:
                return str(value)
            except UnicodeEncodeError:
                return unicode(value).encode('utf-8')

        def _debug(self, message, *args):
            if getattr(self.params, 'disco_debug', False):
                self._disco_message(message % args)

        def _error(self, message, *args):
            import traceback
            trace = traceback.format_exc(15)
            self._disco_message('%s %s' % (message, trace))

        def _disco_message(self, message):
            import disco.util
            disco.util.msg(message)

    go = KeysetReduce(params_)
    return go(iter_)



from inferno.lib.peekable import peekable


class SortedIterator(object):

    def __init__(self, inputs):
        self.collection = [peekable(input) for input in inputs]
        self.collection = sorted(self.collection, key=self._key)

    def __iter__(self):
        return self

    def next(self):
        removes = []
        reinsert = None
        rval = None
        for stream in self.collection:
            try:
                rval = stream.next()
                reinsert = stream
                break
            except StopIteration:
                removes.append(stream)

        if rval:
            for remove in removes:
                self.collection.remove(remove)
            if reinsert:
                self.collection.remove(reinsert)
                try:
                    reinsert.peek()
                except:
                    pass
                else:
                    removes = []
                    reinsert_index = 0
                    for stream in self.collection:
                        try:
                            stream.peek()
                            if self._key(reinsert) < self._key(stream):
                                break
                        except:
                            removes.append(stream)
                        reinsert_index += 1
                    self.collection.insert(reinsert_index, reinsert)
                    for remove in removes:
                        self.collection.remove(remove)
            return rval
        raise StopIteration

    def _key(self, stream):
        try:
            key, value = stream.peek()
            return tuple(key)
        except StopIteration:
            return tuple()



import logging
import os
import threading

from multiprocessing import Pipe
from multiprocessing import Process

import inferno.lib.daemon

from inferno.lib.http_server import launch_server
from inferno.lib.job_factory import JobFactory


log = logging.getLogger(__name__)


class DiscoBall(threading.Thread):

    def __init__(self, instance, port, *args, **kwargs):
        threading.Thread.__init__(self, *args, **kwargs)
        self.instance = instance
        self.port = port
        cwd = os.path.abspath(os.path.dirname(__file__))
        self.base_path = os.path.abspath("%s/.." % cwd)
        self.stopped = False

    def spin(self):
        self.start()

    def jobinfo(self, job_id):
        settings = self.instance.settings
        disco, _ = JobFactory.get_disco_ddfs(settings)
        return disco.jobinfo(job_id)

    # The following methods are RPC style that get invoked through the pipe
    # interface to the Tornado process.

    def get_jobinfo(self, job_id):
        return self.jobinfo(job_id)

    def set_history(self, job_id, job):
        self.instance.history[job_id] = job

    def get_modules(self):
        mods = {}
        for mod, rules in self.instance.rules.iteritems():
            mods[mod] = [rule.name for rule in rules]
        return mods

    def get_rules(self, module_name):
        return self.instance.rules.get(module_name)

    def get_rule_summary(self, module_name, rule_name):
        return self.instance.get_rule_named(module_name, rule_name)

    def run_job(self, module_name, rule_name, params):
        rule = self.instance.get_rule_named(module_name, rule_name)
        if rule:
            return self.instance.run_rule(rule, False, params)
        else:
            return None

    def get_job_histories(self):
        return self.instance.history.values()

    def get_job_history(self, job_id):
        return self.instance.history.get(job_id)

    def run(self, *args, **kwargs):
        try:
            log.info('starting the disco ball....')
            parent, kid = Pipe()
            pipe = inferno.lib.daemon.pickle_connection(kid)
            self.server = Process(target=launch_server,
                    args=(self.base_path, self.port, pipe))
            self.server.daemon = True
            self.server.start()

            while not self.stopped:
                msg = args = None
                try:
                    if parent.poll(5):
                        msg, args = parent.recv()
                except Exception as err:
                    error_msg = "Error receiving message from Tornado: %s"
                    log.error(error_msg, err)
                if msg:
                    try:
                        result = getattr(self, msg)(*args)
                    except Exception as er:
                        error_msg = "Error executing RPC for message: %s. %s"
                        log.error(error_msg, msg, er)
                        parent.send(None)
                    else:
                        try:
                            parent.send(result)
                        except Exception as e:
                            error_msg = "Error executing service %s. %s"
                            log.error(error_msg, msg, e)

        except Exception as e:
            error_msg = "Unable to start DiscoBall on port %s for job %s. %s"
            log.error(error_msg, self.port, self.name, e)



import logging
import os
import pickle
import time

from multiprocessing import Pipe
from multiprocessing.reduction import reduce_connection
from threading import RLock
from multiprocessing.process import Process

from inferno.lib.pid import DaemonPid
from inferno.lib.job_factory import JobFactory
from inferno.lib.lookup_rules import get_rules
from inferno.lib.disco_ball import DiscoBall
from inferno.lib.lookup_rules import get_rule_dict
from inferno.lib.lookup_rules import get_rules_by_name


log = logging.getLogger(__name__)


def pickle_connection(connection):
    return pickle.dumps(reduce_connection(connection))


def unpickle_connection(pickled_connection):
    (func, args) = pickle.loads(pickled_connection)
    return func(*args)


def run_rule_async(rule_name, automatic_cycle, settings, reply_to):
    import sys

    pipe = unpickle_connection(reply_to)
    response_sent = False
    pid_created = False
    parent = 'http://127.0.0.1:%d' % settings.get('inferno_http_port', 6970)
    rules = get_rules_by_name(
        rule_name, settings['rules_directory'], immediate=not automatic_cycle)
    if rules and len(rules) > 0:
        rule = rules[0]
    else:
        log.error('No rule exists with rule_name: %s' % rule_name)
        raise Exception('No rule exists with rule_name: %s' % rule_name)
    job = JobFactory.create_job(rule, settings, parent)
    pid = DaemonPid(settings)

    try:
        if automatic_cycle and pid.should_run(job):
            pid_created = pid.create_pid(job)
            pid.create_last_run(job)

        if not automatic_cycle or pid_created:
            if job.start():
                pipe.send({'job': job.job_msg})
                pipe.close()
                response_sent = True
                job.wait()
            else:
                pipe.send({'error': "job didn't start"})
                pipe.close()
                response_sent = True
        else:
            pipe.send({'warn': "no pid"})
            pipe.close()
            response_sent = True
    except Exception as e:
        if not response_sent:
            pipe.send({'error': e.message})
        log.error('Error running job %s: %s',
                  job.rule_name, e, exc_info=sys.exc_info())
    finally:
        if pid_created:
            pid.remove_pid(job)
        os._exit(0)


class InfernoDaemon(object):

    def __init__(self, settings):
        self.lock = RLock()
        self.settings = settings
        self._paused = settings.get('start_paused')
        self._stopped = False
        self._rules = get_rule_dict(settings['rules_directory'], True)
        self.history = {}

    @property
    def port(self):
        return self.settings.get('inferno_http_port', 6970)

    @property
    def rules(self):
        with self.lock:
            return self._rules

    @property
    def paused(self):
        with self.lock:
            return self._paused

    @property
    def stopped(self):
        with self.lock:
            return self._stopped

    def get_rule_named(self, mod, rule_name):
        with self.lock:
            for rule in self._rules.get(mod, []):
                if rule.name == rule_name:
                    return rule

    def run_rule(self, rule, automatic_cycle=False, params=None):
        try:
            job_settings = self.settings.copy()
            if params:
                job_settings.update(params)
            parent, kid = Pipe()
            reply_to = pickle_connection(kid)
            name = rule.qualified_name
            args = (name, automatic_cycle, job_settings, reply_to)
            Process(target=run_rule_async, args=args).start()
            msg = parent.recv()
            if msg and 'job' in msg:
                job = msg['job']
                self.history[job['job_name']] = job
                return msg['job']
            elif 'error' in msg:
                log.error('Error creating job: %s' % msg)
                return None
        except Exception as e:
            log.error("Error running rule: %s" % e)
            raise e

    def die(self, x=None, y=None):
        print 'dying...'
        try:
            self.disco_ball.server.terminate()
        except:
            pass
        os._exit(0)

    def start(self):
        import signal

        signal.signal(signal.SIGTERM, self.die)

        log.info('Starting Inferno... (pid=%s)', os.getpid())
        auto_rules = get_rules(self.settings['rules_directory'])

        port = self.settings.get('inferno_http_port', 6970)
        self.disco_ball = DiscoBall(instance=self, port=port)
        self.disco_ball.spin()

        # keep cycling through the automatic rules
        while not self.stopped:

            # cycle through all the automatic rules
            for rule in auto_rules:

                if self.stopped:
                    break

                # skip this rule
                if not rule.run or self.paused:
                    continue

                self.run_rule(rule, automatic_cycle=True)

                # sleep for a bit between rules
                time.sleep(self.settings['spawn_delay'])

        self.die()



import copy
import imp
import logging
import os


log = logging.getLogger(__name__)


def get_rules_by_name(name, rules_directory, immediate=False):

    def _get_rule_module(rules_dir, path, mod_name):
        try:
            file, path, desc = imp.find_module(mod_name, [rules_dir])
            mod = imp.load_module(mod_name, file, path, desc)
            return mod
        except Exception as e:
            import traceback
            trace = traceback.format_exc(15)
            log.warning("%s %s" % (e, trace))
            return None

    def _rule_with_requested_keyset(keyset_name, rule):
        rule_copy = copy.deepcopy(rule)
        for keyset_key in rule_copy.params.keysets.keys():
            if keyset_key != keyset_name:
                del rule_copy.params.keysets[keyset_key]
        return rule_copy

    path = name.split('.')
    mod_name = path[0] if len(path) > 0 else None
    rule_name = path[1] if len(path) > 1 else None
    keyset_name = path[2] if len(path) > 2 else None

    mod = _get_rule_module(rules_directory, path, mod_name)
    if not immediate and mod and not getattr(mod, 'AUTORUN', False):
        log.debug('Not immediate mode, skipping rules for %r', name)
        return []

    rules = getattr(mod, 'RULES', [])
    if rule_name and keyset_name:
        for rule in rules:
            if rule.name == rule_name and keyset_name in rule.params.keysets:
                rule_copy = _rule_with_requested_keyset(keyset_name, rule)
                rule_copy.qualified_name = '%s.%s' % (mod, rule.name)
                return [rule_copy]
    elif rule_name:
        for rule in rules:
            if rule.name == rule_name:
                rule.qualified_name = '%s.%s' % (mod_name, rule.name)
                return [rule]
    elif mod and rules:
        return rules

    log.warning('Unable to find rules for %r', name)
    return []


def get_rule_modules(rules_directory):
    modules = []
    for name in os.listdir(rules_directory):
        if name.startswith('.'):
            continue
        name, ext = os.path.splitext(name)
        if ext == '.py' and name != '__init__':
            modules.append(name)
    return modules


def get_rules(rules_directory, immediate=False):
    rules = []
    for mod in get_rule_modules(rules_directory):
        for rule in get_rules_by_name(mod, rules_directory, immediate):
            rule.qualified_name = '%s.%s' % (mod, rule.name)
            rules.append(rule)
    return rules


def get_rule_dict(rules_directory, immediate=False):
    rules = {}
    for mod in get_rule_modules(rules_directory):
        for rule in get_rules_by_name(mod, rules_directory, immediate):
            rule.qualified_name = '%s.%s' % (mod, rule.name)
            mod_list = rules.setdefault(mod, [])
            mod_list.append(rule)
    return rules



import httplib
import logging
import os
import StringIO

from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop

import tornado.web
import ujson
import yaml

import inferno.lib.daemon


log = logging.getLogger(__name__)


class RequestHandler(tornado.web.RequestHandler):

    def get_error_html(self, status_code, **kwargs):
        return {
            "code": status_code,
            "message": kwargs.get('message') or httplib.responses[status_code],
            }

    def build_url(self, *args, **kwargs):
        url = "%s://%s" % (self.request.protocol, self.request.host)
        path = self.request.path

        if kwargs.get('relative'):
            if self.request.path.endswith('.json'):
                path = self.request.path[0:-5]
            url = "%s%s" % (url, path)

        for arg in args:
            url += '/' + arg

        if kwargs.get('json', True):
            return "%s.json" % url
        else:
            return url

    def request_app(self, msg, *args):
        if getattr(self.application, 'unit_testing', True):
            func = getattr(self.application.disco_ball, msg)
            return func(*args)
        try:
            reply = None
            pipe = self.application.pipe
            try:
                pipe.send((msg, args))
            except Exception as e:
                self.send_error("Comm. error: %s" % e)
            else:
                reply = pipe.recv()
            return reply
        except Exception as e:
            self.send_error('Error receiving message: %s' % e)
            return None

    def get_jobinfo(self, job_id):
        try:
            return self.request_app('get_jobinfo', job_id)
        except Exception:
            return None


class MainHandler(RequestHandler):

    def get(self, *args, **kwargs):
        self.write(dict(
            rules=self.build_url('rules', relative=False),
            jobs=self.build_url('jobs', relative=False),
            stop=self.build_url('stop', relative=False, json=False),
            pause=self.build_url('pause', relative=False, json=False),
            play=self.build_url('play', relative=False, json=False),
        ))


class ChildUpdateHandler(RequestHandler):

    def post(self, *args, **kwargs):
        msg = self.request.arguments['msg'][0]
        job = ujson.loads(msg)
        job_id = args[0]
        self.request_app('set_history', job_id, job)


class RuleListHandler(RequestHandler):

    def get(self, *args, **kwargs):
        mods = self.request_app('get_modules')
        rval = {}
        for mod_name, rules in mods.iteritems():
            urls = [self.build_url(mod_name, rule, relative=True)
                    for rule in rules]
            rval[mod_name] = urls
        self.write(rval)


class ModuleHandler(RequestHandler):

    def get(self, *args, **kwargs):
        module = args[0]
        rules = self.request_app('get_rules', module)
        if rules:
            urls = [self.build_url(rule.name, relative=True) for rule in rules]
            self.write({module: urls})
        else:
            self.send_error(message='could not find rules')


class RuleHandler(RequestHandler):

    def get(self, *args, **kwargs):
        rule = self.request_app('get_rule_summary', *args)
        if rule:
            self.write(rule.summary_dict())
        else:
            self.send_error(message='could not find rule')

    def post(self, *args, **kwargs):
        params = {}
        for _, ymls in self.request.arguments.iteritems():
            for yml in ymls:
                params.update(yaml.load(yml))

        job = self.request_app('run_job', args[0], args[1], params)
        if job and 'job_name' in job:
            url = self.build_url('jobs', job['job_name'], relative=False)
            self.write(dict(job_url=url))
        else:
            message = 'could not start job for rule: %s.%s'
            self.send_error(message=message % args)


class JobHandler(RequestHandler):

    def get_job_detail(self, job):
        results = self.build_url('results.csv', relative=True, json=False)
        rule_segs = job['rule_name'].replace('.', '/')
        rule = self.build_url('rules', rule_segs, relative=False)
        rval = job.copy()
        rval['results'] = results
        rval['rule'] = rule
        return rval

    def get_job_summary(self, job):
        url = self.build_url(job['job_name'], relative=True)
        return dict(
            id=job['job_name'],
            status=job['current_stage'],
            url=url
        )

    def get(self, *args, **kwargs):
        if not len(args):
            # list all jobs and their URLs
            job_list = []
            jobs = self.request_app('get_job_histories')
            #for job in self.application.instance.history.itervalues():
            for job in jobs:
                job_list.append(self.get_job_summary(job))
            self.write(dict(jobs=job_list))
        else:
            # list specific details of job
            #job = self.application.instance.history[args[0]]
            job = self.request_app('get_job_history', args[0])
            if job:
                self.write(self.get_job_detail(job))
            else:
                # we don't have the full inferno job info for this job, but
                # maybe disco knows about it...
                jobinfo = self.get_jobinfo(args[0])
                if jobinfo:
                    self.write(dict(jobinfo=jobinfo))
                else:
                    message = "can't retrieve job info for job_id: %s"
                    self.send_error(message=message % args[0])


class JobResultHandler(RequestHandler):

    def get(self, *args, **kwargs):
        from inferno.lib.disco_ext import sorted_iterator
        from inferno.lib.result import keyset_result

        def flush_callback(stringio):
            self.write(stringio.getvalue())
            return StringIO.StringIO()

        self.set_header('Content-Type', 'text/csv; charset=UTF-8')
        job_id = args[0]
        jobinfo = self.get_jobinfo(job_id)
        if jobinfo:
            keyset_result(
                sorted_iterator(jobinfo['results']),
                params=None,
                output_stream=StringIO.StringIO(),
                flush_callback=flush_callback,
                generate_header=False,
                skip_keyset=True)


MAPPINGS = [
    (r"/index.json", MainHandler),
    (r"/rules.json", RuleListHandler),
    (r"/rules/([^/?]*).json", ModuleHandler),
    (r"/rules/([^/?]*)/([^/?]*).json", RuleHandler),
    (r"/jobs.json", JobHandler),
    (r"/jobs/([^/?]*).json", JobHandler),
    (r"/jobs/([^/?]*)/results.csv", JobResultHandler),
    (r"/_status/([^/?]*)", ChildUpdateHandler)
]


def launch_server(base_path, port, pipe):
    path = os.path.join(base_path, "disco_ball", "templates")
    application = tornado.web.Application(MAPPINGS, template_path=path)
    application.pipe = inferno.lib.daemon.unpickle_connection(pipe)
    sockets = tornado.netutil.bind_sockets(port)
    server = HTTPServer(application)
    server.add_sockets(sockets)
    IOLoop.instance().start()



from nose.tools import assert_raises
from nose.tools import eq_

from inferno.lib.peekable import peekable


class TestPeekable(object):

    def test_get_one(self):
        p = peekable(range(4))
        eq_(p.peek(), 0)
        eq_(p.next(), 0)

    def test_get_list(self):
        p = peekable(range(4))
        eq_(p.peek(1), [0])
        eq_(p.next(1), [0])

    def test_look_ahead_then_grab_some(self):
        p = peekable(range(4))
        eq_(p.peek(3), [0, 1, 2])
        eq_(p.next(2), [0, 1])

    def test_walking_off_the_end(self):
        p = peekable(range(100))
        assert_raises(StopIteration, p.peek, 101)
        eq_(p.peek(1), [0])
        assert_raises(StopIteration, p.next, 101)
        eq_(p.next(1), [0])

    def test_iter(self):
        actual = [x for x in peekable(range(4))]
        eq_(actual, [0, 1, 2, 3])



import types

from nose.tools import eq_
from nose.tools import ok_

from disco.core import Params

from inferno.lib.reduce import keyset_reduce


class TestKeysetReduce(object):

    def test_reduce(self):
        data = [
            ('["keyset_BBB", "key1", "key2"]', [1, 10]),
            ('["keyset_BBB", "key1", "key2"]', [2, 20]),
            ('["keyset_ZZZ", "key1", "key2"]', [1, 100]),
            ('["keyset_ZZZ", "key1", "key2"]', [2, 200]),
            ('["keyset_AAA", "key1", "key2"]', [1, 1000]),
            ('["keyset_AAA", "key1", "key2"]', [2, 2000])]
        expected = [
            (['keyset_BBB', 'key1', 'key2'], [3, 30]),
            (['keyset_ZZZ', 'key1', 'key2'], [3, 300]),
            (['keyset_AAA', 'key1', 'key2'], [3, 3000])]
        self._assert_reduce(data, expected)

    def test_reduce_assumes_input_keys_are_grouped_together(self):
        data = [
            ('["keyset_BBB", "key1", "key2"]', [1, 10]),
            ('["keyset_ZZZ", "key1", "key2"]', [1, 100]),
            ('["keyset_AAA", "key1", "key2"]', [1, 1000]),
            ('["keyset_BBB", "key1", "key2"]', [2, 20]),
            ('["keyset_ZZZ", "key1", "key2"]', [2, 200]),
            ('["keyset_AAA", "key1", "key2"]', [2, 2000])]
        expected = [
            (['keyset_BBB', 'key1', 'key2'], [1, 10]),
            (['keyset_ZZZ', 'key1', 'key2'], [1, 100]),
            (['keyset_AAA', 'key1', 'key2'], [1, 1000]),
            (['keyset_BBB', 'key1', 'key2'], [2, 20]),
            (['keyset_ZZZ', 'key1', 'key2'], [2, 200]),
            (['keyset_AAA', 'key1', 'key2'], [2, 2000])]
        # expected equals input b/c the input keys weren't grouped together
        self._assert_reduce(data, expected)

    def test_null_value(self):
        data = [
            ('["keyset", "key1", "key2"]', [1, None])]
        expected = [
            (['keyset', 'key1', 'key2'], [1, 0])]
        self._assert_reduce(data, expected)

    def test_float_value(self):
        data = [
            ('["keyset", "key1", "key2"]', [1, 1.234])]
        expected = [
            (['keyset', 'key1', 'key2'], [1, 1.234])]
        self._assert_reduce(data, expected)

    def test_long_value(self):
        data = [
            ('["keyset", "key1", "key2"]', [1, 1326664799000])]
        expected = [
            (['keyset', 'key1', 'key2'], [1, 1326664799000])]
        self._assert_reduce(data, expected)

    def test_string_value(self):
        data = [
            ('["keyset", "key1", "key2"]', [1, 'some_string'])]
        expected = [
            (['keyset', 'key1', 'key2'], [1, 0])]
        self._assert_reduce(data, expected)

    def test_should_throw_away_bad_data_and_continue(self):
        data = [
            ('["keyset", "key1", "key2"]', [1, 1]),
            ('key_not_json', [1, 2]),
            ('["keyset", "key1", "key2"]', [1, 3])]
        expected = [
            (['keyset', 'key1', 'key2'], [1, 1]),
            (['keyset', 'key1', 'key2'], [1, 3])]
        self._assert_reduce(data, expected)

    def _assert_reduce(self, data, expected):
        # turn disco_debug on for more code coverage
        params = Params(disco_debug=True)
        actual = keyset_reduce(data, params)
        ok_(isinstance(actual, types.GeneratorType))
        eq_(list(actual), expected)



# -*- coding: utf-8 -*-

import cStringIO
import os
import sys
import tempfile

from nose.tools import eq_

from inferno.lib.disco_ext import get_disco_handle
from inferno.lib.job_factory import JobFactory
from inferno.lib.settings import InfernoSettings


class TestExamples(object):

    json_data = """
{"first_name":"Joan", "last_name":"Términos"}
{"first_name":"Willow", "last_name":"Harvey"}
{"first_name":"Noam", "last_name":"Clarke"}
{"first_name":"Joan", "last_name":"Harvey"}
{"first_name":"Beatty", "last_name":"Clarke"}
"""
    csv_data = """
Joan,Términos
Willow,Harvey
Noam,Clarke
Joan,Harvey
Beatty,Clarke
"""

    def setUp(self):
        sys.stdout = self.capture_stdout = cStringIO.StringIO()
        here = os.path.dirname(__file__)
        self.rules_dir = os.path.join(here, '..', 'fixture', 'test_rules')
        self.tag = 'test:integration:chunk:users:2012-01-10'
        self.custom_tag_prefix = 'result:test:integration:last_names_result'
        self.result_tag_prefix = 'disco:job:results:last_names_result'

    def tearDown(self):
        sys.stdout = sys.__stdout__
        self.ddfs.delete(self.tag)
        for prefix in [self.custom_tag_prefix, self.result_tag_prefix]:
            for tag in self.ddfs.list(prefix):
                self.ddfs.delete(tag)

    def test_json(self):
        settings = InfernoSettings(
            rules_directory=self.rules_dir,
            immediate_rule='names.last_names_json')
        expected = [
            'last_name,count',
            'Clarke,2',
            'Harvey,2',
            'Términos,1']
        self._assert_integration_test(settings, expected, self.json_data)

    def test_csv(self):
        settings = InfernoSettings(
            rules_directory=self.rules_dir,
            immediate_rule='names.last_names_csv')
        expected = [
            'last_name,count',
            'Clarke,2',
            'Harvey,2',
            'Términos,1']
        self._assert_integration_test(settings, expected, self.csv_data)

    def test_many_keysets(self):
        settings = InfernoSettings(
            rules_directory=self.rules_dir,
            immediate_rule='names.first_and_last_names')
        expected = [
            'first_name,count',
            'Beatty,1',
            'Joan,2',
            'Noam,1',
            'Willow,1',
            'last_name,count',
            'Clarke,2',
            'Harvey,2',
            'Términos,1']
        self._assert_integration_test(settings, expected, self.json_data)

    def test_tag_results(self):
        settings = InfernoSettings(
            rules_directory=self.rules_dir,
            immediate_rule='names.last_names_result')
        expected = [
            'last_name,count',
            'Clarke,2',
            'Harvey,2',
            'Términos,1']
        self._assert_integration_test(settings, expected, self.json_data)
        tags = self.ddfs.list(self.custom_tag_prefix)
        eq_(len(tags), 1)
        tags = self.ddfs.list(self.result_tag_prefix)
        eq_(len(tags), 1)

    def _chunk_test_data(self, settings, data):
        (_, name) = tempfile.mkstemp()
        with open(name, 'w') as f:
            f.write(data)
        self.settings = settings
        (_, self.ddfs) = get_disco_handle(settings['server'])
        self.ddfs.delete(self.tag)
        self.ddfs.chunk(self.tag, [name])

    def _assert_integration_test(self, settings, expected, data):
        self._chunk_test_data(settings, data)
        JobFactory.execute_immediate_rule(settings)
        eq_(self.capture_stdout.getvalue(), '\r\n'.join(expected) + '\r\n')



def keyset_multiplier(stream, size, url, params):
    for in_dict in stream:
        for keyset_name in params.keysets.keys():
            dprime = dict(in_dict)
            dprime['_keyset'] = keyset_name
            yield dprime


def json_reader(stream, size=None, url=None, params=None):
    import ujson
    for line in stream:
        if line.find('{') != -1:
            try:
                parts = ujson.loads(line)
            except:
                # just skip bad lines
                import disco.util
                disco.util.msg('json line error: %r' % line)
                pass
            else:
                yield parts


def csv_reader(stream, size=None, url=None, params=None):
    import csv
    fieldnames = params.csv_fields
    dialect = params.csv_dialect
    for line in stream:
        try:
            reader = csv.DictReader(
                [line], fieldnames=fieldnames, dialect=dialect)
            for parts in reader:
                if None in parts:
                    # remove extra data values
                    del parts[None]
                yield parts
        except:
            # just skip bad lines
            import disco.util
            disco.util.msg('csv line error: %r' % line)
            pass



import random

from nose.tools import eq_

from inferno.lib.lazy_property import lazy_property


class TestLazyProperty(object):

    @lazy_property
    def not_so_random(self):
        return random.randint(0, 100)

    def test_lazy_property(self):
        value1 = self.not_so_random
        value2 = self.not_so_random
        eq_(value1, value2)



import logging.config
import sys

import argparse
import yaml

from datetime import date
from time import mktime
from time import strptime

from inferno.lib import __version__
from inferno.lib.job_factory import JobFactory
from inferno.lib.settings import InfernoSettings


def _get_options(argv):
    desc = 'Inferno: a python map/reduce platform powered by disco.'
    parser = argparse.ArgumentParser(description=desc)

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version='%s-%s' % ('inferno', __version__))

    parser.add_argument(
        "-s",
        "--server",
        dest="server",
        help="master disco server")

    parser.add_argument(
        "-e",
        "--settings",
        dest="settings_file",
        help="path to settings file")

    parser.add_argument(
        "-i",
        "--immediate-rule",
        dest="immediate_rule",
        help="execute <module>.<rule> immediately and exit")

    parser.add_argument(
        "-y",
        "--rules-directory",
        dest="rules_directory",
        help="directory to search for Inferno rules")

    parser.add_argument(
        "-f",
        "--force",
        dest="force",
        default=False,
        action="store_true",
        help="force processing of blobs")

    parser.add_argument(
        "-x",
        "--start-paused",
        dest="start_paused",
        default=False,
        action="store_true",
        help="start Inferno without starting any automatic rules (pause mode)")

    parser.add_argument(
        "-D",
        "--disco-debug",
        dest="disco_debug",
        default=False,
        action="store_true",
        help="debug map/reduce to disco console")

    parser.add_argument(
        "-d",
        "--debug",
        dest="debug",
        default=False,
        action="store_true",
        help="debug flag for inferno consumers")

    parser.add_argument(
        "-p",
        "--profile",
        dest="profile",
        default=False,
        action="store_true",
        help="output disco profiling data")

    parser.add_argument(
        "-t",
        "--source-tags",
        dest="source_tags",
        help="override the ddfs source tags")

    parser.add_argument(
        "-r",
        "--result-tag",
        dest="result_tag",
        help="override the ddfs result tag")

    parser.add_argument(
        "-S",
        "--day-start",
        dest="day_start",
        help="override the start day for blobs")

    parser.add_argument(
        "-R",
        "--day-range",
        dest="day_range",
        type=int,
        help="override the day range for blobs")

    parser.add_argument(
        "-O",
        "--day-offset",
        dest="day_offset",
        type=int,
        help="override the days previous to start day for blobs")

    parser.add_argument(
        "-P",
        "--parameters",
        dest="parameters",
        default=[],
        action="append",
        help="additional rule parameters (in yaml)")

    options = parser.parse_args(argv)

    if options.source_tags:
        options.source_tags = options.source_tags.split(',')

    if options.day_start:
        options.day_start = date.fromtimestamp(mktime(strptime(
            options.day_start, '%Y-%m-%d')))

    result = options.__dict__
    for parameter in options.parameters:
        result.update(yaml.load(parameter))

    return result


def _get_settings(options):
    if options['settings_file']:
        settings = InfernoSettings(settings_file=options['settings_file'])
    else:
        settings = InfernoSettings()
    for key, value in options.iteritems():
        if value is not None or key not in settings:
            settings[key] = value
    return settings


def main(argv=sys.argv):
    options = _get_options(argv[1:])
    settings = _get_settings(options)
    log = logging.getLogger(__name__)

    try:
        log_config = settings.get('log_config')
        logging.config.fileConfig(log_config, disable_existing_loggers=False)
    except Exception as e:
        f = '%(asctime)s,%(msecs)03d %(levelname)-5.5s [%(name)s] %(message)s'
        logging.basicConfig(level=logging.INFO, format=f)
        log.error('Error setting up logging [%s]: %s' % (log_config, e))

    log.info('Starting inferno-%s with: \n%r', __version__, settings)

    for path in settings.get('extra_python_paths'):
        sys.path.insert(0, path)

    # either run inferno in 'immediate' or 'daemon' mode
    if options['immediate_rule']:
        JobFactory.execute_immediate_rule(settings)
    else:
        from inferno.lib.daemon import InfernoDaemon
        InfernoDaemon(settings).start()



import csv
import os
import StringIO
import types

from nose.tools import eq_
from nose.tools import ok_
from nose.tools import raises

from disco.core import Params

from inferno.lib.reader import csv_reader
from inferno.lib.reader import json_reader
from inferno.lib.reader import keyset_multiplier


class TestKeysetMultiplier(object):

    def test_keyset_multiplier(self):
        params = Params()
        params.keysets = {
            'last_name_keyset': dict(
                key_parts=['_keyset', 'last_name'],
                value_parts=['count'],
             ),
            'first_name_keyset': dict(
                key_parts=['_keyset', 'first_name'],
                value_parts=['count'],
             )}
        data = [
            {'first_name': 'Willow', 'last_name': 'Harvey'},
            {'first_name': 'Noam', 'last_name': 'Clarke'}]
        expected = [
            {
                'first_name': 'Willow',
                'last_name': 'Harvey',
                '_keyset': 'first_name_keyset'
            },
            {
                'first_name': 'Willow',
                'last_name': 'Harvey',
                '_keyset': 'last_name_keyset'
            },
            {
                'first_name': 'Noam',
                'last_name': 'Clarke',
                '_keyset': 'first_name_keyset'
            },
            {
                'first_name': 'Noam',
                'last_name': 'Clarke',
                '_keyset': 'last_name_keyset'
            }]
        actual = keyset_multiplier(data, None, None, params)
        ok_(isinstance(actual, types.GeneratorType))
        eq_(list(actual), expected)


class TestJsonReader(object):

    def test_json_reader(self):
        self._assert_json_reader('input.json')

    def test_json_with_empty_lines(self):
        self._assert_json_reader('input_empty_lines.json')

    def test_json_with_bad_lines(self):
        self._assert_json_reader('input_bad_lines.json')

    @raises(Exception)
    def test_bad_stream(self):
        list(json_reader(10000, None, None, None))

    def _assert_json_reader(self, name):
        here = os.path.dirname(__file__)
        path = os.path.join(here, '..', 'fixture', name)
        actual = json_reader(open(path), None, None, None)
        expected = [
            {
                'timestamp': 1319475603759L,
                'token': '92f80832-f43b-11e0-a45f-00259035c924',
                'is_new': False,
                'id': 10000
            },
            {
                'timestamp': 1319475603987L,
                'token': '9e28ace0-fe61-11e0-97da-00259006bc9c',
                'is_new': True,
                'id': 20000
            }
        ]
        ok_(isinstance(actual, types.GeneratorType))
        eq_(list(actual), expected)


class TestCSVReader(object):

    def test_csv_reader(self):
        fields = ('field_1', 'field_2', 'field_3')
        values = 'value_1	value_2	value_3'
        expected = [{
            'field_1': 'value_1',
            'field_2': 'value_2',
            'field_3': 'value_3'}]
        self._assert_csv_reader(fields, values, expected)

    def test_more_fields_than_input(self):
        fields = ('field_1', 'field_2', 'field_3')
        values = 'value_1	value_2'
        expected = [{
            'field_1': 'value_1',
            'field_2': 'value_2',
            'field_3': None}]
        self._assert_csv_reader(fields, values, expected)

    def test_less_fields_than_input(self):
        fields = ('field_1', 'field_2', 'field_3')
        values = 'value_1	value_2	value_3	value_4'
        expected = [{
            'field_1': 'value_1',
            'field_2': 'value_2',
            'field_3': 'value_3'}]
        self._assert_csv_reader(fields, values, expected)

    def test_empty_value_in_the_middle(self):
        fields = ('field_1', 'field_2', 'field_3')
        values = 'value_1		value_3'
        expected = [{
            'field_1': 'value_1',
            'field_2': '',
            'field_3': 'value_3'}]
        self._assert_csv_reader(fields, values, expected)

    def test_empty_values_on_each_side(self):
        fields = ('field_1', 'field_2', 'field_3')
        values = '	value_2	'
        expected = [{
            'field_1': '',
            'field_2': 'value_2',
            'field_3': ''}]
        self._assert_csv_reader(fields, values, expected)

    def _assert_csv_reader(self, fields, values, expected):
        stream = StringIO.StringIO(values)
        params = Params()
        params.csv_fields = fields
        params.csv_dialect = csv.excel_tab
        actual = csv_reader(stream, None, None, params)
        ok_(isinstance(actual, types.GeneratorType))
        eq_(list(actual), expected)



import datetime
import types

from nose.tools import eq_
from nose.tools import ok_

from inferno.lib.map import keyset_map
from inferno.lib.rule import InfernoRule


class TestKeysetMap(object):

    def setUp(self):
        self.data = {
            'city': 'toronto',
            'country': 'canada',
            'population': 100,
            'size': 1000,
            'date': datetime.date(2012, 12, 01)}
        self.rule = InfernoRule(
            key_parts=['country', 'city'],
            value_parts=['population', 'size'])

    def test_no_key_parts(self):
        rule = InfernoRule(
            key_parts=[],
            value_parts=['population', 'size'])
        eq_(rule.params.keysets, {})
        expected = []
        self._assert_map(self.data, rule, expected)

    def test_no_value_parts(self):
        rule = InfernoRule(
            key_parts=['country', 'city'],
            value_parts=[])
        eq_(rule.params.keysets, {})
        expected = []
        self._assert_map(self.data, rule, expected)

    def test_keys_and_parts(self):
        expected = [('[null,"canada","toronto"]', [100, 1000])]
        self._assert_map(self.data, self.rule, expected)

    def test_missing_key_part_should_not_yield_result(self):
        del self.data['city']
        expected = []
        self._assert_map(self.data, self.rule, expected)

    def test_missing_value_part_should_yield_result(self):
        del self.data['size']
        expected = [('[null,"canada","toronto"]', [100, 0])]
        self._assert_map(self.data, self.rule, expected)

    def test_null_key_part_should_not_yield_result(self):
        self.data['city'] = None
        expected = []
        self._assert_map(self.data, self.rule, expected)

    def test_null_value_part_should_yield_result(self):
        self.data['size'] = None
        expected = [('[null,"canada","toronto"]', [100, None])]
        self._assert_map(self.data, self.rule, expected)

    def test_empty_key_part_should_yield_result(self):
        self.data['city'] = ''
        expected = [('[null,"canada",""]', [100, 1000])]
        self._assert_map(self.data, self.rule, expected)

    def test_empty_value_part_should_yield_result(self):
        self.data['size'] = ''
        expected = [('[null,"canada","toronto"]', [100, ''])]
        self._assert_map(self.data, self.rule, expected)

    def test_map_serialization(self):
        # key parts are str casted & json serialized, value parts are are not
        # (note the difference between the key date and value date results)
        rule = InfernoRule(
            key_parts=['date'],
            value_parts=['date'])
        expected = [('[null,"2012-12-01"]', [datetime.date(2012, 12, 1)])]
        self._assert_map(self.data, rule, expected)

    def test_field_transforms(self):
        def upper(val):
            return val.upper()

        rule = InfernoRule(
            key_parts=['country', 'city'],
            value_parts=['population', 'size'],
            field_transforms={'city': upper, 'country': upper})
        expected = [('[null,"CANADA","TORONTO"]', [100, 1000])]
        self._assert_map(self.data, rule, expected)

    def test_parts_preprocess_that_yields_multiple_parts(self):
        def lookup_language(parts, params):
            for language in ['french', 'english']:
                parts_copy = parts.copy()
                parts_copy['language'] = language
                yield parts_copy

        rule = InfernoRule(
            key_parts=['country'],
            value_parts=['language'],
            parts_preprocess=[lookup_language])
        expected = [
            ('[null,"canada"]', ['french']),
            ('[null,"canada"]', ['english'])]
        self._assert_map(self.data, rule, expected)

    def test_field_transforms_happen_after_parts_preprocess(self):
        def lookup_language(parts, params):
            for language in ['french', 'english']:
                parts_copy = parts.copy()
                parts_copy['language'] = language
                yield parts_copy

        def upper(val):
            return val.upper()

        rule = InfernoRule(
            key_parts=['country'],
            value_parts=['language'],
            parts_preprocess=[lookup_language],
            field_transforms={'language': upper})
        expected = [
            ('[null,"canada"]', ['FRENCH']),
            ('[null,"canada"]', ['ENGLISH'])]
        self._assert_map(self.data, rule, expected)

    def _assert_map(self, parts, rule, expected):
        # turn disco_debug on for more code coverage
        rule.params.disco_debug = True
        actual = keyset_map(parts, rule.params)
        ok_(isinstance(actual, types.GeneratorType))
        eq_(list(actual), expected)



import types

from datetime import date

from mock import Mock
from nose.tools import eq_
from nose.tools import ok_

from disco.core import Params

from inferno.lib.job import JOB_ARCHIVE
from inferno.lib.job import JOB_BLOBS
from inferno.lib.job import JOB_PROCESS
from inferno.lib.job import JOB_PROFILE
from inferno.lib.job import JOB_PURGE
from inferno.lib.job import JOB_RESULTS
from inferno.lib.job import JOB_TAG
from inferno.lib.job import InfernoJob
from inferno.lib.rule import InfernoRule
from inferno.lib.settings import InfernoSettings

from test.mock.disco import DDFS
from test.mock.disco import Disco


class TestJob(object):

    MAX_BLOBS = 1000
    ARCHIVE_PREFIX = 'archived'

    def setUp(self):
        settings = InfernoSettings(day_range=2, day_start=date(2011, 11, 12))
        rule = InfernoRule(
            archive=True,
            max_blobs=self.MAX_BLOBS,
            name='some_rule_name',
            archive_tag_prefix='archived',
            source_tags=['incoming:data:chunk'])
        self.job = InfernoJob(rule, settings, Params())
        self.job.disco = Disco()
        self.job.ddfs = DDFS()

    def test_start_not_enough_blobs(self):
        self.job.rule.min_blobs = 1000
        self.job.start()
        eq_(self.job.current_stage, JOB_BLOBS)

    def test_determine_job_blobs(self):
        expected_tags = [
            'incoming:data:chunk:2011-11-12',
            'incoming:data:chunk:2011-11-11']
        expected_blobs = [
            ('b12.1', 'b12.2', 'b12.3'),
            ('b11.1', 'b11.2', 'b11.3')]
        self.job.current_stage = None
        archiver = self.job._determine_job_blobs()

        # check that the archiver was created correctly
        eq_(archiver.max_blobs, self.MAX_BLOBS)
        eq_(archiver.archive_mode, True)
        eq_(archiver.archive_prefix, self.ARCHIVE_PREFIX)

        # check that it found the correct tags and blobs
        eq_(archiver.tags, expected_tags)
        eq_(archiver.job_blobs, expected_blobs)

        # check that the state changed
        eq_(self.job.current_stage, JOB_BLOBS)

    def test_archive_tags(self):
        # there should be no archived tags before calling archive
        archive_prefix = 'archived:data:chunk'
        archiver = self.job._determine_job_blobs()
        self.job.current_stage = None
        archived = archiver.ddfs.list(archive_prefix)
        eq_(archived, [])

        # should not archive & change state since archive mode is false
        archiver.archive_mode = False
        self.job._archive_tags(archiver)
        eq_(self.job.current_stage, None)
        archived = archiver.ddfs.list(archive_prefix)
        eq_(archived, [])

        # should archive & change state since archive mode is true
        archiver.archive_mode = True
        self.job._archive_tags(archiver)
        eq_(self.job.current_stage, JOB_ARCHIVE)
        archived = archiver.ddfs.list(archive_prefix)
        expected = [
            'archived:data:chunk:2011-11-11',
            'archived:data:chunk:2011-11-12']
        eq_(archived, expected)

    def test_get_job_results_no_results(self):
        self.job.current_stage = None
        urls = []
        actual = self.job._get_job_results(urls)
        ok_(isinstance(actual, types.GeneratorType))
        eq_(list(actual), [])
        eq_(self.job.current_stage, JOB_RESULTS)

    def test_process_results_no_results(self):
        self.job.current_stage = None
        results = []
        self.job._process_results(results, 'some_job_id')
        eq_(self.job.current_stage, JOB_PROCESS)

    def test_purge(self):
        # should purge if there's no 'no_purge' setting
        ok_('no_purge' not in self.job.settings)
        self.job.current_stage = None
        self.job._purge('some_job_name')
        eq_(self.job.current_stage, JOB_PURGE)

        # should not purge & change state
        self.job.settings['no_purge'] = True
        self.job.current_stage = None
        self.job._purge('some_job_name')
        eq_(self.job.current_stage, None)

        # should purge & change state
        self.job.settings['no_purge'] = False
        self.job.current_stage = None
        self.job._purge('some_job_name')
        eq_(self.job.current_stage, JOB_PURGE)

    def test_profile(self):
        # should not profile if there's no 'profile' setting
        ok_('profile' not in self.job.settings)
        self.job.current_stage = None
        self.job._profile(Mock())
        eq_(self.job.current_stage, None)

        # should not profile & change state
        self.job.settings['profle'] = False
        self.job.current_stage = None
        self.job._profile(Mock())
        eq_(self.job.current_stage, None)

        # should profile & change state
        self.job.settings['profile'] = True
        self.job.current_stage = None
        self.job._profile(Mock())
        eq_(self.job.current_stage, JOB_PROFILE)

    def test_tag_results(self):
        # should not tag results & change state
        self.job.settings['result_tag'] = None
        self.job.current_stage = None
        self.job._tag_results('some_job_name')
        eq_(self.job.current_stage, None)

        # should tag results & change state
        before = self.job.ddfs.list('some_result_tag')
        eq_(before, [])
        self.job.settings['result_tag'] = 'some_result_tag'
        self.job.current_stage = None
        self.job._tag_results('some_job_name')
        eq_(self.job.current_stage, JOB_TAG)
        after = self.job.ddfs.list('some_result_tag')
        eq_(after, ['some_result_tag:some_job_name'])

    def test_update_state(self):
        self.job.current_stage = None
        self.job._update_state('some_state')
        eq_(self.job.current_stage, 'some_state')

    def test_enough_blobs(self):
        # equal
        self.job.rule.min_blobs = 100
        self.job.settings['force'] = False
        eq_(self.job._enough_blobs(blob_count=100), True)

        # more than
        self.job.rule.min_blobs = 100
        self.job.settings['force'] = False
        eq_(self.job._enough_blobs(blob_count=101), True)

        # less than
        self.job.rule.min_blobs = 100
        self.job.settings['force'] = False
        eq_(self.job._enough_blobs(blob_count=99), False)

        # less than, but force mode is enabled
        self.job.rule.min_blobs = 100
        self.job.settings['force'] = True
        eq_(self.job._enough_blobs(blob_count=99), True)

    def test_str(self):
        eq_(str(self.job), '<InfernoJob for: some_rule_name>')



from nose.tools import eq_

from inferno.lib.sorted_iterator import SortedIterator


class TestSortedIterator(object):

    def test_merges_sorted_inputs(self):
        data = [
            [
                ((1, 1), 'some_value'),
                ((1, 2), 'some_value'),
                ((1, 3), 'some_value')
            ],
            [
                ((1, 100), 'some_value'),
                ((1, 200), 'some_value'),
                ((1, 300), 'some_value')
            ],
            [
                ((1, 10), 'some_value'),
                ((1, 20), 'some_value'),
                ((1, 30), 'some_value')
            ],
            [
                ((1, 4), 'some_value'),
                ((1, 40), 'some_value'),
                ((1, 400), 'some_value')
            ]
        ]
        sorted_iterator = SortedIterator(data)
        expected = [
            ((1, 1), 'some_value'),
            ((1, 2), 'some_value'),
            ((1, 3), 'some_value'),
            ((1, 4), 'some_value'),
            ((1, 10), 'some_value'),
            ((1, 20), 'some_value'),
            ((1, 30), 'some_value'),
            ((1, 40), 'some_value'),
            ((1, 100), 'some_value'),
            ((1, 200), 'some_value'),
            ((1, 300), 'some_value'),
            ((1, 400), 'some_value')]
        eq_(list(sorted_iterator), expected)

    def test_assumes_individual_inputs_are_already_sorted(self):
        data = [
            [
                ((2, 1), 'some_value'),
                ((1, 1), 'some_value'),
            ],
            [
                ((4, 1), 'some_value'),
                ((3, 1), 'some_value'),
            ]
        ]
        sorted_iterator = SortedIterator(data)
        expected = [
            ((2, 1), 'some_value'),
            ((1, 1), 'some_value'),
            ((4, 1), 'some_value'),
            ((3, 1), 'some_value')]
        eq_(list(sorted_iterator), expected)

    def test_handles_duplicates(self):
        data = [
            [
                ((1, 1), 'some_value'),
                ((1, 2), 'some_value'),
            ],
            [
                ((1, 1), 'some_value'),
                ((1, 2), 'some_value'),
                ((1, 3), 'some_value'),
            ],
            [
                ((1, 3), 'some_value'),
            ]
        ]
        sorted_iterator = SortedIterator(data)
        expected = [
            ((1, 1), 'some_value'),
            ((1, 1), 'some_value'),
            ((1, 2), 'some_value'),
            ((1, 2), 'some_value'),
            ((1, 3), 'some_value'),
            ((1, 3), 'some_value')]
        eq_(list(sorted_iterator), expected)

    def test_handles_empty_input(self):
        data = [
            [((1, 1), 'some_value')],
            [],  # <----- empty input
            [((2, 1), 'some_value')],
        ]
        sorted_iterator = SortedIterator(data)
        expected = [
            ((1, 1), 'some_value'),
            ((2, 1), 'some_value')]
        eq_(list(sorted_iterator), expected)



import os

import yaml


def guess_settings():
    for settings_file in (
        os.path.expanduser('~/.inferno'),
        '/etc/inferno/settings.yaml'):
        if os.path.exists(settings_file):
            return settings_file
    return ''


defaults = {
    'log_config': '/etc/inferno/log.ini',
    'settings_file': guess_settings(),
    'spawn_delay': 5,
    'max_workers': 8,
    'pid_dir': '/var/run/inferno',
    'server': 'localhost',
    'rules_directory': '/apps/project/project/rules',
    'extra_python_paths': [],
}


class InfernoSettings(dict):

    def __init__(self, *args, **kwargs):
        # load the defaults
        super(InfernoSettings, self).update(defaults)

        # override with the settings file
        path = kwargs.get('settings_file') or self['settings_file']
        if path and os.path.exists(path):
            self.update(yaml.load(open(path)))

        # final overrides
        super(InfernoSettings, self).__init__(*args, **kwargs)



from inferno.lib.disco_ext import get_disco_handle
from inferno.lib.job import InfernoJob
from inferno.lib.lookup_rules import get_rules_by_name


class JobFactory(object):

    @staticmethod
    def get_immediate_rules(settings):
        immed_rule = settings.get('immediate_rule')
        rules_dir = settings.get('rules_directory')
        return get_rules_by_name(immed_rule, rules_dir, immediate=True)

    @staticmethod
    def execute_immediate_rule(settings, parent=None):
        rules = JobFactory.get_immediate_rules(settings)
        for rule in rules:
            job = JobFactory.create_job(rule, settings, parent)
            if job.start():
                job.wait()

    @staticmethod
    def get_disco_ddfs(settings):
        server = settings.get('server')
        return get_disco_handle(server)

    @staticmethod
    def create_job(rule, settings, parent=None):
        return InfernoJob(rule, settings, parent)



import os
import shutil
import tempfile

from disco.worker.classic.worker import Params

from nose.tools import assert_raises
from nose.tools import raises
from nose.tools import eq_
from nose.tools import ok_

from inferno.lib.job import InfernoJob
from inferno.lib.pid import DaemonPid
from inferno.lib.rule import InfernoRule
from inferno.lib.settings import InfernoSettings


class TestDaemonPid(object):

    def setUp(self):
        self.settings = InfernoSettings()
        self._make_temp_pid_dir()
        self.job = InfernoJob(InfernoRule(name='some_rule_name'), {}, Params())
        self.pid = DaemonPid(self.settings)

    def test_create_pid(self):
        # create the pid file
        eq_(self.pid.create_pid(self.job), True)
        # can't create it again
        eq_(self.pid.create_pid(self.job), False)

    @raises(OSError)
    def test_create_pid_exception(self):
        unknown_path = os.path.join('some', 'unknown', 'path')
        self.pid._settings['pid_dir'] = unknown_path
        self.pid.create_pid(self.job)

    def test_remove_pid(self):
        # create the pid file
        eq_(self.pid.create_pid(self.job), True)
        # remove the pid file
        self.pid.remove_pid(self.job)
        # can't remove it again
        assert_raises(OSError, self.pid.remove_pid, self.job)

    def test_processes(self):
        eq_(self.pid.processes, [])
        eq_(self.pid.create_pid(self.job), True)
        actual = self.pid.processes
        eq_(len(actual), 1)
        eq_(set(actual[0].keys()), set(['pid', 'timestamp', 'name']))
        eq_(actual[0]['name'], 'some_rule_name')
        ok_(int(actual[0]['pid']) > 0)
        eq_(int(actual[0]['pid']), os.getpid())

    def test_should_run(self):
        path = self.pid._last_run_path(self.job)
        # without last run file
        eq_(os.path.exists(path), False)
        eq_(self.pid.should_run(self.job), True)
        # with last run file that's new
        self.pid.create_last_run(self.job)
        eq_(self.pid.should_run(self.job), False)
        # with last run file that's old
        with open(path, 'w') as f:
            f.write('2005-12-01 12:34:56')
        eq_(self.pid.should_run(self.job), True)

    def _make_temp_pid_dir(self):
        temp_dir = tempfile.gettempdir()
        self.settings['pid_dir'] = os.path.join(temp_dir, 'inferno')
        pid_dir = self.settings['pid_dir']
        if os.path.exists(pid_dir):
            shutil.rmtree(pid_dir)



import json
import os
import tempfile

from mock import Mock
from mock import patch
from nose.tools import eq_

import disco.core
import tornado.testing
import tornado.web

from inferno.lib.daemon import InfernoDaemon
from inferno.lib.disco_ball import DiscoBall
from inferno.lib.http_server import MAPPINGS
from inferno.lib.job_factory import JobFactory


FAKE_JOB_ID = 'automatic_rule_1@532:4f2b9:f0d86'


class DiscoBallHandlerTestCase(tornado.testing.AsyncHTTPTestCase):

    def get_app(self):
        self.application = tornado.web.Application(MAPPINGS)
        self.application.disco_ball = DiscoBall(FakeInfernoDaemon(), None)
        self.application.unit_testing = True
        return self.application

    def get_path(self, port):
        return 'http://localhost:%s' % (port)

    def put_a_fake_job_in_the_history(self, path):
        url = '/rules/automatic/automatic_rule_1.json'
        expected = {'job_url': '%s/jobs/%s.json' % (path, FAKE_JOB_ID)}
        self.assert_json_response_to_post(url, expected)

    def assert_csv_response(self, url, expected, code=200):
        response = self.fetch(url)
        content_type = 'text/csv; charset=UTF-8'
        eq_(response.headers['Content-Type'], content_type)
        eq_(response.body, expected)
        eq_(response.code, code)

    def assert_json_response_to_post(self, url, expected, code=200):
        response = self.fetch(url, method='POST', body='')
        self._assert_json_response(expected, code, response)

    def assert_json_response(self, url, expected, code=200):
        response = self.fetch(url)
        self._assert_json_response(expected, code, response)

    def _assert_json_response(self, expected, code, response):
        content_type = 'application/json; charset=UTF-8'
        eq_(response.headers['Content-Type'], content_type)
        eq_(json.loads(response.body), expected)
        eq_(response.code, code)


class TestMainHandler(DiscoBallHandlerTestCase):

    def test_get(self):
        path = self.get_path(self.get_http_port())
        url = '/index.json'
        expected = {
            'stop': '%s/stop' % path,
            'play': '%s/play' % path,
            'pause': '%s/pause' % path,
            'jobs': '%s/jobs.json' % path,
            'rules': '%s/rules.json' % path}
        self.assert_json_response(url, expected)


class TestRuleHandler(DiscoBallHandlerTestCase):

    def test_post(self):
        path = self.get_path(self.get_http_port())
        url = '/rules/automatic/automatic_rule_1.json'
        expected = {'job_url': '%s/jobs/%s.json' % (path, FAKE_JOB_ID)}
        self.assert_json_response_to_post(url, expected)

    def test_post_unknown_rule(self):
        url = '/rules/automatic/unknown.json'
        expected = {
            'message': 'could not start job for rule: automatic.unknown',
            'code': 500}
        self.assert_json_response_to_post(url, expected, code=500)

    def test_get(self):
        url = '/rules/automatic/automatic_rule_1.json'
        expected = {
            'name': 'automatic_rule_1',
            'run': True,
            'archive': False,
            'map_input_stream': [],
            'map_function': 'inferno.lib.map.keyset_map',
            'reduce_function': 'inferno.lib.reduce.keyset_reduce',
            'parts_preprocess': [],
            'keysets': {
                'keyset_1': {
                    'key_parts': ['_keyset', 'key_1'],
                    'table': None,
                    'value_parts': ['value_1'],
                    'column_mappings': None
                },
                'keyset_2': {
                    'key_parts': ['_keyset', 'key_2'],
                    'value_parts': ['value_2'],
                    'table': None,
                    'column_mappings': None
                },
            },
        }
        self.assert_json_response(url, expected)

    def test_get_unknown_rule(self):
        url = '/rules/automatic/unknown.json'
        expected = {'message': 'could not find rule', 'code': 500}
        self.assert_json_response(url, expected, code=500)


class TestModuleHandler(DiscoBallHandlerTestCase):

    def test_get(self):
        path = self.get_path(self.get_http_port())
        url = '/rules/automatic.json'
        expected = {
            'automatic': [
                '%s/rules/automatic/automatic_rule_1.json' % path,
                '%s/rules/automatic/automatic_rule_2.json' % path,
                '%s/rules/automatic/automatic_rule_3.json' % path]}
        self.assert_json_response(url, expected)

    def test_get_unknown_module(self):
        url = '/rules/unknown.json'
        expected = {'message': 'could not find rules', 'code': 500}
        self.assert_json_response(url, expected, code=500)


class TestRuleListHandler(DiscoBallHandlerTestCase):

    def test_get(self):
        path = self.get_path(self.get_http_port())
        url = '/rules.json'
        expected = {
            'automatic': [
                '%s/rules/automatic/automatic_rule_1.json' % path,
                '%s/rules/automatic/automatic_rule_2.json' % path,
                '%s/rules/automatic/automatic_rule_3.json' % path],
            'more_automatic': [
                '%s/rules/more_automatic/automatic_rule_4.json' % path],
            'manual': [
                '%s/rules/manual/manual_rule_1.json' % path,
                '%s/rules/manual/manual_rule_2.json' % path,
                '%s/rules/manual/manual_rule_3.json' % path],
            'more_manual': [
                '%s/rules/more_manual/manual_rule_4.json' % path]}
        self.assert_json_response(url, expected)


class TestJobHandler(DiscoBallHandlerTestCase):

    def test_get_job_list(self):
        path = self.get_path(self.get_http_port())
        self.put_a_fake_job_in_the_history(path)
        url = '/jobs.json'
        expected = {
            'jobs': [{
                'url': '%s/jobs/%s.json' % (path, FAKE_JOB_ID),
                'status': 'job.start',
                'id': FAKE_JOB_ID}]}
        self.assert_json_response(url, expected)

    def test_get_empty_job_list(self):
        url = '/jobs.json'
        expected = {'jobs': []}
        self.assert_json_response(url, expected)

    def test_get_specific_job(self):
        path = self.get_path(self.get_http_port())
        self.put_a_fake_job_in_the_history(path)
        url = '/jobs/%s.json' % FAKE_JOB_ID
        settings = self.application.disco_ball.instance.settings
        expected = {
            'settings': settings,
            'child_data': None,
            'current_stage': 'job.start',
            'rule_name': 'automatic.automatic_rule_1',
            'job_name': 'automatic_rule_1@532:4f2b9:f0d86',
            'rule': '%s/rules/automatic/automatic_rule_1.json' % path,
            'results': '%s/jobs/%s/results.csv' % (path, FAKE_JOB_ID),
        }
        self.assert_json_response(url, expected)


class TestJobResultHandler(DiscoBallHandlerTestCase):

    @patch.object(disco.core.Disco, 'jobinfo')
    @patch('inferno.lib.disco_ext.sorted_iterator')
    def test_get(self, mock_sorted_iterator, mock_jobinfo):
        mock_jobinfo.return_value = {'results': ['url_1', 'url_2']}
        mock_sorted_iterator.return_value = [
            (['keyset_1', 'Tim'], [100]),
            (['keyset_1', 'Tom'], [200]),
            (['keyset_1', 'Sam'], [300])]
        path = self.get_path(self.get_http_port())
        self.put_a_fake_job_in_the_history(path)
        url = '/jobs/%s/results.csv' % FAKE_JOB_ID
        expected = 'Tim,100\r\nTom,200\r\nSam,300\r\n'
        self.assert_csv_response(url, expected)

    @patch.object(disco.core.Disco, 'jobinfo')
    def test_get_empty_results(self, mock_jobinfo):
        mock_jobinfo.return_value = {'results': []}
        path = self.get_path(self.get_http_port())
        self.put_a_fake_job_in_the_history(path)
        url = '/jobs/%s/results.csv' % FAKE_JOB_ID
        expected = ''
        self.assert_csv_response(url, expected)


class FakeInfernoDaemon(InfernoDaemon):

    def __init__(self):
        here = os.path.dirname(__file__)
        rules_dir = os.path.join(here, '..', 'fixture', 'fake_rules')
        settings = {
            'rules_directory': rules_dir,
            'pid_dir': tempfile.gettempdir()}
        InfernoDaemon.__init__(self, settings)

    def run_rule(self, rule, *args, **kwargs):
        # just override and fake it
        job = JobFactory.create_job(rule, self.settings)
        job.job = Mock()
        job.job.name = FAKE_JOB_ID
        self.history[FAKE_JOB_ID] = job.job_msg
        return job.job_msg



from inferno.lib.rule import InfernoRule
from inferno.lib.rule import Keyset


RULES = [
    InfernoRule(
        name='manual_rule_4',
        keysets={
            'keyset_1':Keyset(
                key_parts=['key_1'],
                value_parts=['value_1'],
             ),
            'keyset_2':Keyset(
                key_parts=['key_2'],
                value_parts=['value_2']
             )
        }
    ),
]



import sys

from datetime import date

from disco.core import Params
from disco.core import result_iterator
from disco.func import chain_stream
from disco.func import disco_output_stream
from disco.func import gzip_stream
from disco.func import map_output_stream
from disco.func import reduce_output_stream

from inferno.lib.disco_ext import sorted_iterator
from inferno.lib.map import keyset_map
from inferno.lib.reader import csv_reader
from inferno.lib.reader import json_reader
from inferno.lib.reader import keyset_multiplier
from inferno.lib.reduce import keyset_reduce
from inferno.lib.result import keyset_result


gzip_csv_stream = gzip_stream + (csv_reader,)
gzip_json_stream = gzip_stream + (json_reader,)
chunk_json_stream = chain_stream + (json_reader,)
chunk_json_keyset_stream = chain_stream + (json_reader, keyset_multiplier)
chunk_csv_stream = chain_stream + (csv_reader,)
chunk_csv_keyset_stream = chain_stream + (csv_reader, keyset_multiplier)


class Keyset(object):

    def __init__(self,
                 key_parts=None,
                 value_parts=None,
                 column_mappings=None,
                 table=None):

        self.key_parts = ['_keyset'] + list(key_parts)
        self.value_parts = value_parts
        self.column_mappings = column_mappings
        self.table = table

    def as_dict(self):
        return {
            'key_parts': self.key_parts,
            'value_parts': self.value_parts,
            'column_mappings': self.column_mappings,
            'table': self.table}


class InfernoRule(object):

    def __init__(self,
                 # name, on/off
                 name='_unnamed_',
                 run=True,

                 # throttle
                 min_blobs=1,
                 max_blobs=sys.maxint,
                 partitions=200,
                 scheduler=None,
                 time_delta=None,

                 # archive
                 archive=False,
                 archive_tag_prefix='processed',

                 # map
                 map_init_function=lambda x, y: x,
                 map_function=keyset_map,
                 map_input_stream=None,
                 map_output_stream=(
                     map_output_stream, disco_output_stream),

                 # reduce
                 reduce_function=keyset_reduce,
                 reduce_output_stream=(
                     reduce_output_stream, disco_output_stream),

                 # result
                 result_processor=keyset_result,
                 result_tag=None,
                 save=False,
                 sort=True,

                 # keysets
                 keysets=None,
                 key_parts=None,
                 value_parts=None,
                 column_mappings=None,
                 table=None,

                 # input
                 day_range=0,
                 day_offset=0,
                 day_start=date.today(),
                 source_tags=None,

                 # other
                 rule_init_function=None,
                 parts_preprocess=None,
                 field_transforms=None,
                 **kwargs):

        self.qualified_name = name
        if kwargs:
            self.params = Params(**kwargs)
        else:
            self.params = Params()

        if not scheduler:
            scheduler = {'force_local': True, 'max_cores': 200}

        # name, on/off
        self.run = run
        self.name = name

        # throttle
        self.min_blobs = min_blobs
        self.max_blobs = max_blobs
        self.partitions = partitions
        self.scheduler = scheduler
        self.time_delta = time_delta
        if self.time_delta is None:
            self.time_delta = {'minutes': 5}

        # archive
        self.archive = archive
        self.archive_tag_prefix = archive_tag_prefix

        # map
        self.map_init_function = map_init_function
        self.map_function = map_function
        self.map_input_stream = map_input_stream
        self.map_output_stream = map_output_stream

        # reduce
        self.reduce_function = reduce_function
        self.reduce_output_stream = reduce_output_stream

        # result
        self.result_processor = result_processor
        self.result_tag = result_tag
        self.save = save
        self.sort = sort
        if self.sort:
            self.result_iterator = sorted_iterator
        else:
            self.result_iterator = result_iterator

        # input
        if isinstance(source_tags, basestring):
            source_tags = [source_tags]
        self.day_range = day_range
        self.day_offset = day_offset
        self.day_start = day_start
        self.source_tags = source_tags

        # keysets
        keyset_dict = {}
        if keysets:
            for keyset_name, keyset_obj in keysets.items():
                keyset_dict[keyset_name] = keyset_obj.as_dict()
        elif key_parts and value_parts:
            keyset_dict['_default'] = Keyset(
                key_parts,
                value_parts,
                column_mappings,
                table).as_dict()
        self.params.keysets = keyset_dict

        # preprocess
        if parts_preprocess:
            self.params.parts_preprocess = map(
                lambda func: func.__name__, parts_preprocess)
            for func in parts_preprocess:
                self.params.__setattr__(func.__name__, func)
        else:
            self.params.parts_preprocess = []

        # transforms
        if field_transforms:
            self.params.field_transforms = {}
            for key, func in field_transforms.items():
                self.params.field_transforms[key] = func.__name__
                self.params.__setattr__(func.__name__, func)

        # other
        self.rule_init_function = rule_init_function

    def __str__(self):
        return '<InfernoRule: %s>' % self.name

    @property
    def is_atomic(self):
        return self.archive

    def summary_dict(self):
        def fstr(func):
            return "%s.%s" % (func.__module__, func.__name__)

        def fname(funcvec):
            # serialize collections of funcs as collections of func names
            return [fstr(func) for func in funcvec] if funcvec else []

        # mostly just a dump of the __dict__, some exceptions
        return dict(
            name=self.name,
            run=self.run,
            archive=self.archive,
            map_input_stream=fname(self.map_input_stream),
            map_function=fstr(self.map_function),
            reduce_function=fstr(self.reduce_function),
            keysets=self.params.keysets,
            parts_preprocess=self.params.parts_preprocess)



def keyset_result(iter, params, **kwargs):

    import csv
    import sys

    def default_flush_callback(stream):
        stream.flush()
        return stream

    output_stream = kwargs.get('output_stream', sys.stdout)
    flush_schedule = kwargs.get('flush_schedule', 5000)
    flush_callback = kwargs.get('flush_callback', default_flush_callback)
    generate_header = kwargs.get('generate_header', True)
    skip_keyset = kwargs.get('skip_keyset', True)

    writer = csv.writer(output_stream)
    last_keyset_name = None
    count = 0
    for keys, values in iter:
        keyset_name = keys[0]
        if generate_header and last_keyset_name != keyset_name:
            columns = params.keysets[keyset_name]['key_parts'][1:]
            columns.extend(params.keysets[keyset_name]['value_parts'])
            writer.writerow(columns)
            last_keyset_name = keyset_name

        if skip_keyset:
            data = keys[1:] + values
        else:
            data = keys + values

        row = [unicode(x).encode('utf-8') for x in data]
        writer.writerow(row)
        count += 1
        if count > flush_schedule:
            output_stream = flush_callback(output_stream)
            count = 0
    flush_callback(output_stream)



import os

from nose.tools import eq_

from inferno.lib.lookup_rules import get_rules
from inferno.lib.lookup_rules import get_rule_modules
from inferno.lib.lookup_rules import get_rules_by_name


class TestGetRulesByName(object):

    def setUp(self):
        here = os.path.dirname(__file__)
        self.rules_dir = os.path.join(here, '..', 'fixture', 'fake_rules')

    def test_empty_name(self):
        self._assert_get_rules_by_name('', [])

    def test_unknown_module(self):
        name = 'some_unknown_module'
        self._assert_get_rules_by_name(name, [])

    def test_unknown_rule(self):
        name = 'automatic.some_unknown_rule'
        self._assert_get_rules_by_name(name, [])

    def test_unknown_keyset(self):
        name = 'automatic.automatic_rule_1.some_unknown_keyset'
        self._assert_get_rules_by_name(name, [])

    def test_automatic_module(self):
        name = 'automatic'
        expected = [
            ['automatic_rule_1', ['keyset_1', 'keyset_2']],
            ['automatic_rule_2', ['keyset_1', 'keyset_2']],
            ['automatic_rule_3', ['_default']]]
        self._assert_get_rules_by_name(name, expected)
        self._assert_get_rules_by_name(name, expected, immediate=False)
        self._assert_get_rules_by_name(name, expected, immediate=True)

    def test_automatic_module_rule(self):
        name = 'automatic.automatic_rule_1'
        expected = [['automatic_rule_1', ['keyset_1', 'keyset_2']]]
        self._assert_get_rules_by_name(name, expected)
        self._assert_get_rules_by_name(name, expected, immediate=False)
        self._assert_get_rules_by_name(name, expected, immediate=True)

    def test_automatic_module_rule_keyset(self):
        name = 'automatic.automatic_rule_1.keyset_2'
        expected = [['automatic_rule_1', ['keyset_2']]]
        self._assert_get_rules_by_name(name, expected)
        self._assert_get_rules_by_name(name, expected, immediate=False)
        self._assert_get_rules_by_name(name, expected, immediate=True)

    def test_manual_module(self):
        name = 'manual'
        expected = [
            ['manual_rule_1', ['keyset_1', 'keyset_2']],
            ['manual_rule_2', ['keyset_1', 'keyset_2']],
            ['manual_rule_3', ['_default']]]
        self._assert_get_rules_by_name(name, [])
        self._assert_get_rules_by_name(name, [], immediate=False)
        self._assert_get_rules_by_name(name, expected, immediate=True)

    def test_manual_module_rule(self):
        name = 'manual.manual_rule_1'
        expected = [['manual_rule_1', ['keyset_1', 'keyset_2']]]
        self._assert_get_rules_by_name(name, [])
        self._assert_get_rules_by_name(name, [], immediate=False)
        self._assert_get_rules_by_name(name, expected, immediate=True)

    def test_manual_module_rule_keyset(self):
        name = 'manual.manual_rule_1.keyset_2'
        expected = [['manual_rule_1', ['keyset_2']]]
        self._assert_get_rules_by_name(name, [])
        self._assert_get_rules_by_name(name, [], immediate=False)
        self._assert_get_rules_by_name(name, expected, immediate=True)

    def _assert_get_rules_by_name(self, name, expected, immediate=None):
        if immediate is None:
            rules = get_rules_by_name(name, self.rules_dir)
        else:
            rules = get_rules_by_name(name, self.rules_dir, immediate)
        actual = [[rule.name, rule.params.keysets.keys()] for rule in rules]
        for rule in actual:
            rule[1].sort()
        eq_(actual, expected)


class TestLookupRules(object):

    def setUp(self):
        here = os.path.dirname(__file__)
        self.rules_dir = os.path.join(here, '..', 'fixture', 'fake_rules')

    def test_get_rule_modules(self):
        actual = get_rule_modules(self.rules_dir)
        expected = [
            'automatic',
            'manual',
            'more_automatic',
            'more_manual']
        eq_(actual, expected)

    def test_get_automatic_rules(self):
        rules = get_rules(self.rules_dir)
        actual = [rule.name for rule in rules]
        expected = [
            'automatic_rule_1',
            'automatic_rule_2',
            'automatic_rule_3',
            'automatic_rule_4']
        eq_(actual, expected)



import ujson
import logging
import sys
import time
import urllib2
import urllib

from disco.core import Params
from disco.job import Job

from inferno.lib.archiver import Archiver
from inferno.lib.disco_ext import get_disco_handle
from inferno.lib.job_options import JobOptions
from inferno.lib.result import keyset_result


log = logging.getLogger(__name__)

JOB_ARCHIVE = 'job.archive'
JOB_BLOBS = 'job.blobs'
JOB_DONE = 'job.done'
JOB_PROCESS = 'job.process'
JOB_PROFILE = 'job.profile'
JOB_PURGE = 'job.purge'
JOB_RESULTS = 'job.results'
JOB_RUN = 'job.run'
JOB_START = 'job.start'
JOB_TAG = 'job.tag'
JOB_WAIT = 'job.wait'
JOB_ERROR = 'job.error'


class InfernoJob(object):

    def __init__(self, rule, settings, parent=None):
        self.disco, self.ddfs = get_disco_handle(settings.get('server'))
        self.job_options = JobOptions(rule, settings)
        self.rule = rule
        self.settings = settings
        rule_params = dict(rule.params.__dict__)
        rule_params.update(settings)
        self.params = Params(**rule_params)
        self.job = Job(name=rule.name, master=self.disco.master)
        self.full_job_id = None
        self.parent = parent
        self.jobinfo = None
        self._update_state(JOB_START)

    @property
    def job_name(self):
        return self.job.name

    @property
    def rule_name(self):
        return self.rule.qualified_name

    @property
    def job_msg(self):
        return dict(
            job_name=self.job_name,
            rule_name=self.rule_name,
            settings=self.settings,
            child_data=self.jobinfo,
            current_stage=self.current_stage,
        )

    def start(self):
        self.start_time = time.time()
        self.archiver = self._determine_job_blobs()
        if self._enough_blobs(self.archiver.blob_count):
            if self.rule.rule_init_function:
                self.rule.rule_init_function(self.params)
            self.job.run(name=self.rule.name,
                  input=self.archiver.job_blobs,
                  map=self.rule.map_function,
                  reduce=self.rule.reduce_function,
                  params=self.params,
                  partitions=self.rule.partitions,
                  map_input_stream=self.rule.map_input_stream,
                  map_output_stream=self.rule.map_output_stream,
                  map_init=self.rule.map_init_function,
                  save=self.rule.save or self.rule.result_tag is not None,
                  scheduler=self.rule.scheduler,
                  reduce_output_stream=self.rule.reduce_output_stream,
                  sort=self.rule.sort,
                  profile=self.settings.get('profile'))

            # actual id is only assigned after starting the job
            self.full_job_id = self.job.name
            self._update_state(JOB_RUN)
            return self.job
        return None

    def wait(self):
        blob_count = self.archiver.blob_count
        log.info('Started job %s processing %i blobs',
            self.job.name, blob_count)
        self._notify_parent(JOB_WAIT)
        try:
            jobout = self.job.wait()
            log.info('Done waiting for job %s', self.job.name)
            results = self._get_job_results(jobout)
            self._profile(self.job)
            self._tag_results(self.job.name)
            if not self.settings.get('debug'):
                self._process_results(results, self.job.name)
            else:
                keyset_result(results, self.rule.params)
            self._purge(self.job.name)
        except Exception as e:
            log.error('Job %s failed: %s',
                      self.job.name, e, exc_info=sys.exc_info())
            self._notify_parent(JOB_ERROR)
        else:
            if not self.settings.get('debug'):
                self._archive_tags(self.archiver)
            self._notify_parent(JOB_DONE)
        log.info('Finished job %s', self.job.name)

    def _determine_job_blobs(self):
        self._update_state(JOB_BLOBS)
        tags = self.job_options.tags
        log.info('Processing tags: %s', tags)
        archiver = Archiver(
            ddfs=self.ddfs,
            archive_prefix=self.rule.archive_tag_prefix,
            archive_mode=self.rule.archive,
            max_blobs=self.rule.max_blobs,
            tags=tags)
        return archiver

    def _get_job_results(self, jobout):
        if self.rule.result_iterator:
            self._notify_parent(JOB_RESULTS)
            return self.rule.result_iterator(jobout)

    def _profile(self, job):
        if self.settings.get('profile'):
            self._notify_parent(JOB_PROFILE)
            job.profile_stats().sort_stats('cumulative').print_stats()

    def _tag_results(self, job_name):
        if self.job_options.result_tag:
            self._notify_parent(JOB_TAG)
            result_name = 'disco:job:results:%s' % job_name
            tag_name = '%s:%s' % (self.job_options.result_tag, job_name)
            log.info('Tagging result: %s', tag_name)
            try:
                self.ddfs.tag(tag_name, list(self.ddfs.blobs(result_name)))
            except Exception as e:
                log.error('Error tagging result %s: %s',
                    tag_name, e, exc_info=sys.exc_info())

    def _process_results(self, results, job_id):
        if self.rule.result_processor:
            self._notify_parent(JOB_PROCESS)
            self.rule.result_processor(
                results, params=self.rule.params, job_id=job_id)

    def _purge(self, job_name):
        if not self.settings.get('no_purge'):
            self._notify_parent(JOB_PURGE)
            self.disco.purge(job_name)

    def _archive_tags(self, archiver):
        if archiver.archive_mode:
            self._notify_parent(JOB_ARCHIVE)
            archiver.archive()

    def _update_state(self, stage):
        try:
            self.jobinfo = self.job.jobinfo()
        except:
            pass
        self.current_stage = stage

    def _notify_parent(self, stage):
        self._update_state(stage)
        # if we are daemon spawn, tell mommy where we are
        if self.parent and self.full_job_id:
            msg = ujson.dumps(self.job_msg)
            msg = urllib.urlencode([('msg', msg)])
            url = '%s/_status/%s' % (self.parent, self.full_job_id)
            urllib2.urlopen(url, data=msg, timeout=5)

    def _enough_blobs(self, blob_count):
        if not blob_count or (blob_count < self.rule.min_blobs and
                              not self.settings.get('force')):
            log.info('Skipping job %s: %d blobs required, have only %d',
                self.rule.name, self.rule.min_blobs, blob_count)
            return False
        return True

    def __str__(self):
        return '<InfernoJob for: %s>' % self.rule.name



# -*- coding: utf-8 -*-

import cStringIO
import sys

from nose.tools import eq_

from disco.core import Params

from inferno.lib.result import keyset_result


class TestKeysetResult(object):

    def setUp(self):
        sys.stdout = self.capture_stdout = cStringIO.StringIO()
        self.params = Params()
        self.params.keysets = {
            'last_name_keyset': dict(
                key_parts=['_keyset', 'last_name'],
                value_parts=['count'],
             ),
            'first_name_keyset': dict(
                key_parts=['_keyset', 'first_name'],
                value_parts=['count'],
             )}

    def tearDown(self):
        sys.stdout = sys.__stdout__

    def test_utf8(self):
        data = [(['last_name_keyset', unicode('Términos', 'utf-8')], [30])]
        expected = ['last_name,count', 'Términos,30']
        self._assert_keyset_result(data, self.params, expected)

    def test_input_data_keyset_grouped_together(self):
        data = [
            (['first_name_keyset', 'Tim'], [100]),
            (['first_name_keyset', 'Tom'], [200]),
            (['first_name_keyset', 'Sam'], [300]),
            (['last_name_keyset', 'Clarke'], [10]),
            (['last_name_keyset', 'Harvey'], [20]),
            (['last_name_keyset', 'Martin'], [30])]
        expected = [
            'first_name,count',
            'Tim,100',
            'Tom,200',
            'Sam,300',
            'last_name,count',
            'Clarke,10',
            'Harvey,20',
            'Martin,30']
        self._assert_keyset_result(data, self.params, expected)

    def test_input_data_keyset_not_grouped_together(self):
        data = [
            (['first_name_keyset', 'Tim'], [100]),
            (['last_name_keyset', 'Clarke'], [10]),
            (['first_name_keyset', 'Tom'], [200]),
            (['last_name_keyset', 'Harvey'], [20]),
            (['first_name_keyset', 'Sam'], [300]),
            (['last_name_keyset', 'Martin'], [30])]
        expected = [
            'first_name,count',
            'Tim,100',
            'last_name,count',
            'Clarke,10',
            'first_name,count',
            'Tom,200',
            'last_name,count',
            'Harvey,20',
            'first_name,count',
            'Sam,300',
            'last_name,count',
            'Martin,30']
        self._assert_keyset_result(data, self.params, expected)

    def _assert_keyset_result(self, data, params, expected):
        keyset_result(data, params)
        eq_(self.capture_stdout.getvalue(), '\r\n'.join(expected) + '\r\n')



from disco import func
from disco import util
from disco.settings import DiscoSettings

from inferno.lib.sorted_iterator import SortedIterator


def get_disco_handle(server):
    from disco.core import Disco
    from disco.ddfs import DDFS

    if server and not server.startswith('disco://'):
        server = 'disco://' + server

    return Disco(server), DDFS(server)


def sorted_iterator(urls,
                    reader=func.chain_reader,
                    input_stream=(func.map_input_stream,),
                    notifier=func.notifier,
                    params=None,
                    ddfs=None):

    from disco.worker import Input
    from disco.worker.classic.worker import Worker

    worker = Worker(map_reader=reader, map_input_stream=input_stream)
    settings = DiscoSettings(DISCO_MASTER=ddfs) if ddfs else DiscoSettings()

    inputs = []
    for input in util.inputlist(urls, settings=settings):
        notifier(input)
        instream = Input(input, open=worker.opener('map', 'in', params))
        if instream:
            inputs.append(instream)

    sorted_iter = SortedIterator(inputs)
    for item in sorted_iter:
        yield item



from inferno.lib.settings import InfernoSettings
from inferno.lib.daemon import InfernoDaemon


class TestInfernoDaemon(object):

    def setUp(self):
        self.settings = InfernoSettings()
        self.daemon = InfernoDaemon(self.settings)



from functools import wraps


def lazy_property(f):
    @property
    @wraps(f)
    def wrapper(self, *args, **kwargs):
        lazy = '_lazy_' + f.__name__
        try:
            return getattr(self, lazy)
        except AttributeError:
            result = f(self, *args, **kwargs)
            setattr(self, lazy, result)
            return result
    return wrapper



import glob
import logging
import os
import time

from datetime import datetime
from datetime import timedelta


log = logging.getLogger(__name__)


class DaemonPid(object):

    def __init__(self, settings):
        self._settings = settings
        self._format = '%Y-%m-%d %H:%M:%S'

    @property
    def processes(self):
        files = glob.glob(os.path.join(self._pid_dir, '*.pid'))
        result = []
        for pid_file in files:
            with open(pid_file) as f:
                pid = f.read().strip()
                result.append({
                    'pid': pid,
                    'timestamp': time.ctime(os.path.getctime(pid_file)),
                    'name': os.path.basename(pid_file).replace('.pid', ''),
                })
        return result

    def should_run(self, job):
        path = self._last_run_path(job)
        if not os.path.exists(path):
            return True
        with open(path) as f:
            now = datetime.utcnow()
            delta = timedelta(**job.rule.time_delta)
            last_run = datetime.strptime(f.read().strip(), self._format)
            next_run = last_run + delta
            result = next_run < now
            if not result:
                log.info('Skipping job: %s (last: %s, next: %s)',
                    job.rule_name, last_run.strftime(self._format),
                    next_run.strftime(self._format))
            return result

    def create_pid(self, job):
        path = self._get_pid_path(job)
        if os.path.exists(path):
            return False
        with open(path, 'w') as f:
            f.write(str(os.getpid()))
        return True

    def create_last_run(self, job):
        path = self._last_run_path(job)
        with open(path, 'w') as f:
            now = datetime.utcnow()
            f.write(now.strftime(self._format))

    def remove_pid(self, job):
        path = self._get_pid_path(job)
        os.unlink(path)

    def _get_pid_path(self, job):
        return os.path.join(self._pid_dir, '%s.pid' % job.rule_name)

    def _last_run_path(self, job):
        return os.path.join(self._pid_dir, '%s.last_run' % job.rule_name)

    @property
    def _pid_dir(self):
        path = self._settings['pid_dir']
        if not os.path.exists(path):
            os.mkdir(path)
        return path



import os
import re
import sys

from setuptools import setup, find_packages


if sys.version_info[:2] < (2, 6):
    raise RuntimeError('Requires Python 2.6 or better')

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.txt')).read()
info = open(os.path.join(here, 'inferno', 'lib', '__init__.py')).read()
VERSION = re.compile(r".*__version__ = '(.*?)'", re.S).match(info).group(1)

install_requires = [
    'argparse',
    'pyyaml',
    'setproctitle'
]

tests_require = install_requires + [
    'nose',
    'mock']

setup(
    name='inferno',
    version=VERSION,
    description=('Inferno: a python map/reduce platform powered by disco.'),
    long_description=README,
    keywords='inferno discoproject',
    author='Chango Inc.',
    author_email='dev@chango.com',
    url='http://chango.com',
    license='MIT License',
    packages=find_packages(exclude=['test']),
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    tests_require=tests_require,
    test_suite='nose.collector',
    entry_points="""
    [console_scripts]
    inferno=inferno.bin.run:main
    """)



from inferno.lib.rule import chunk_csv_keyset_stream
from inferno.lib.rule import chunk_json_keyset_stream
from inferno.lib.rule import InfernoRule
from inferno.lib.rule import Keyset


def count(parts, params):
    parts['count'] = 1
    yield parts


RULES = [
    InfernoRule(
        name='last_names_json',
        source_tags=['test:integration:chunk:users'],
        map_input_stream=chunk_json_keyset_stream,
        parts_preprocess=[count],
        partitions=2,
        key_parts=['last_name'],
        value_parts=['count'],
    ),
    InfernoRule(
        name='last_names_csv',
        source_tags=['test:integration:chunk:users'],
        map_input_stream=chunk_csv_keyset_stream,
        csv_fields=('first_name', 'last_name'),
        csv_dialect='excel',
        parts_preprocess=[count],
        partitions=2,
        key_parts=['last_name'],
        value_parts=['count'],
    ),
    InfernoRule(
        name='last_names_result',
        source_tags=['test:integration:chunk:users'],
        map_input_stream=chunk_json_keyset_stream,
        parts_preprocess=[count],
        partitions=2,
        key_parts=['last_name'],
        value_parts=['count'],
        result_tag='result:test:integration'
    ),
    InfernoRule(
        name='first_and_last_names',
        source_tags=['test:integration:chunk:users'],
        map_input_stream=chunk_json_keyset_stream,
        parts_preprocess=[count],
        partitions=2,
        keysets={
            'last_name_keyset':Keyset(
                key_parts=['last_name'],
                value_parts=['count'],
             ),
            'first_name_keyset':Keyset(
                key_parts=['first_name'],
                value_parts=['count']
             )
        }
    )
]



import logging
import sys

from inferno.lib.lazy_property import lazy_property


log = logging.getLogger(__name__)


class Archiver(object):

    def __init__(self, ddfs, archive_prefix, archive_mode, max_blobs, tags):
        self.tags = tags
        self.ddfs = ddfs
        self.max_blobs = max_blobs
        self.archive_mode = archive_mode
        self.archive_prefix = archive_prefix
        self.tag_map = self._build_tag_map(tags)

    @lazy_property
    def blob_count(self):
        return len(self.job_blobs)

    @lazy_property
    def job_blobs(self):
        job_blobs = []
        for blobs in self.tag_map.itervalues():
            job_blobs += blobs
        return job_blobs

    def archive(self):
        if self.tag_map:
            try:
                self._archive_tags()
            except Exception as e:
                log.error('Archiving error: %s', e, exc_info=sys.exc_info())
            finally:
                log.info('Archiving complete')

    def _archive_tags(self):
        for tag, blobs in self.tag_map.iteritems():
            if tag.startswith(self.archive_prefix):
                log.info('Tag already starts with archive prefix: %s', tag)
            else:
                archive_name = self._get_archive_name(tag)
                self.ddfs.tag(archive_name, blobs)
                log.info('Archived %d blobs to %s', len(blobs), archive_name)

    def _build_tag_map(self, tags):
        tag_map = {}
        blob_count = 0
        source_tags, archived_blobs = self._source_and_archived_sets(tags)
        for tag in source_tags:
            blobs = [tuple(sorted(b)) for b in self.ddfs.blobs(tag)]
            blobs = set(blobs) - archived_blobs
            if blobs:
                tag_map[tag] = list(blobs)[:(self.max_blobs - blob_count)]
                blob_count += len(blobs)
                if blob_count >= self.max_blobs:
                    break
        return tag_map

    def _source_and_archived_sets(self, tags):
        source_tags = set()
        archived_blobs = set()
        for prefix in tags:
            for tag in self.ddfs.list(prefix):
                source_tags.add(tag)
                if (self.archive_mode and
                    not tag.startswith(self.archive_prefix)):
                    blobs = self.ddfs.blobs(self._get_archive_name(tag))
                    archived_blobs.update([tuple(sorted(b)) for b in blobs])
        source_tags = sorted(source_tags, reverse=True)
        return source_tags, archived_blobs

    def _get_archive_name(self, tag):
        tag_parts = tag.split(':')
        tag_parts[0] = self.archive_prefix
        return ':'.join(tag_parts)



import tempfile

from datetime import date

from nose.tools import eq_

from inferno.bin.run import _get_options
from inferno.bin.run import _get_settings


class TestOptions(object):

    def test_defaults(self):
        options = _get_options([])
        expected = {
            'parameters': [],
            'force': False,
            'profile': False,
            'debug': False,
            'disco_debug': False,
            'server': None,
            'settings_file': None,
            'source_tags': None,
            'result_tag': None,
            'day_start': None,
            'day_range': None,
            'day_offset': None,
            'immediate_rule': None,
            'rules_directory': None,
            'start_paused': False}
        eq_(options, expected)

    def test_force(self):
        self._assert_force('-f')
        self._assert_force('--force')

    def test_profile(self):
        self._assert_profile('-p')
        self._assert_profile('--profile')

    def test_debug(self):
        self._assert_debug('-d')
        self._assert_debug('--debug')

    def test_disco_debug(self):
        self._assert_disco_debug('-D')
        self._assert_disco_debug('--disco-debug')

    def test_server(self):
        self._assert_server('-s')
        self._assert_server('--server')

    def test_settings(self):
        self._assert_settings('-e')
        self._assert_settings('--settings')

    def test_rules_directory(self):
        self._assert_rules_directory('-y')
        self._assert_rules_directory('--rules-directory')

    def test_day_range(self):
        self._assert_day_range('-R')
        self._assert_day_range('--day-range')

    def test_day_offset(self):
        self._assert_day_offset('-O')
        self._assert_day_offset('--day-offset')

    def test_day_start(self):
        self._assert_day_start('-S')
        self._assert_day_start('--day-start')

    def test_source_tags(self):
        self._assert_source_tags('-t')
        self._assert_source_tags('--source-tags')

    def test_result_tag(self):
        self._assert_result_tag('-r')
        self._assert_result_tag('--result-tag')

    def test_immediate_rule(self):
        self._assert_immediate_rule('-i')
        self._assert_immediate_rule('--immediate-rule')

    def test_parameters(self):
        self._assert_parameters('-P')
        self._assert_parameters('--parameter')

    def _assert_force(self, flag):
        options = _get_options([flag])
        eq_(options['force'], True)

    def _assert_profile(self, flag):
        options = _get_options([flag])
        eq_(options['profile'], True)

    def _assert_debug(self, flag):
        options = _get_options([flag])
        eq_(options['debug'], True)

    def _assert_disco_debug(self, flag):
        options = _get_options([flag])
        eq_(options['disco_debug'], True)

    def _assert_settings(self, flag):
        options = _get_options([flag, 'path_to_settings_file'])
        eq_(options['settings_file'], 'path_to_settings_file')

    def _assert_rules_directory(self, flag):
        options = _get_options([flag, 'path_to_rules_directory'])
        eq_(options['rules_directory'], 'path_to_rules_directory')

    def _assert_day_range(self, flag):
        options = _get_options([flag, '10'])
        eq_(options['day_range'], 10)

    def _assert_day_offset(self, flag):
        options = _get_options([flag, '5'])
        eq_(options['day_offset'], 5)

    def _assert_day_start(self, flag):
        options = _get_options([flag, '2012-12-01'])
        eq_(options['day_start'], date(2012, 12, 1))

    def _assert_immediate_rule(self, flag):
        options = _get_options([flag, 'some_module.some_rule'])
        eq_(options['immediate_rule'], 'some_module.some_rule')

    def _assert_result_tag(self, flag):
        options = _get_options([flag, 'some_result_tag'])
        eq_(options['result_tag'], 'some_result_tag')

    def _assert_parameters(self, flag):
        # one param
        options = _get_options([flag, 'some_param: some_value'])
        eq_(options['some_param'], 'some_value')

        # many params
        options = _get_options([
            flag, 'some_param_1: some_value_1',
            flag, 'some_param_2: some_value_2'])
        eq_(options['some_param_1'], 'some_value_1')
        eq_(options['some_param_2'], 'some_value_2')

        # integer
        options = _get_options([flag, 'some_int: 100'])
        eq_(options['some_int'], 100)

        # list of integers
        options = _get_options([flag, 'some_int_list: [100, 200, 300,]'])
        eq_(options['some_int_list'], [100, 200, 300])

    def _assert_source_tags(self, flag):
        # one tag
        options = _get_options([flag, 'tag1'])
        eq_(options['source_tags'], ['tag1'])

        # many tags
        options = _get_options([flag, 'tag1,tag2'])
        eq_(options['source_tags'], ['tag1', 'tag2'])

    def _assert_server(self, flag):
        options = _get_options([flag, 'some_server'])
        eq_(options['server'], 'some_server')


class TestSettings(object):

    def test_get_settings(self):
        # settings default
        options = _get_options(['-e', 'some_unknown_settings_file'])
        settings = _get_settings(options)
        eq_(settings['server'], 'localhost')

        # options override
        options = _get_options(['-s', 'some_server'])
        settings = _get_settings(options)
        eq_(settings['server'], 'some_server')

        # file override
        settings_file = self._create_settings_file()
        options = _get_options(['-e', settings_file])
        settings = _get_settings(options)
        eq_(settings['server'], 'another_server')

        # server from -s trumps -e
        settings_file = self._create_settings_file()
        options = _get_options(['-e', settings_file, '-s', 'some_server'])
        settings = _get_settings(options)
        eq_(settings['server'], 'some_server')

    def _create_settings_file(self):
        (_, settings_file) = tempfile.mkstemp()
        with open(settings_file, 'w') as f:
            f.write("server: another_server\n")
        return settings_file



from nose.tools import eq_
from nose.tools import ok_

from inferno.lib.archiver import Archiver

from test.mock.disco import DDFS


class TestArchiver(object):

    def _setup(self, tags=()):
        self.archiver = Archiver(
            ddfs=DDFS(),
            archive_prefix='processed',
            archive_mode=True,
            max_blobs=100,
            tags=tags)

    def test_get_archive_name(self):
        self._setup()
        tag = 'incoming:data:chunk:2012-12-01'
        actual = self.archiver._get_archive_name(tag)
        eq_(actual, 'processed:data:chunk:2012-12-01')

    def test_blob_count(self):
        self._setup()
        self.archiver.tag_map = self.fake_tag_map
        eq_(self.archiver.blob_count, 5)

    def test_job_blobs(self):
        self._setup()
        self.archiver.tag_map = self.fake_tag_map
        expected = [
            ('blob1.a', 'blob1.b', 'blob1.c'),
            ('blob2.a', 'blob2.b', 'blob2.c'),
            ('blob3.a', 'blob3.b', 'blob3.c'),
            ('blob4.a', 'blob4.b', 'blob4.c'),
            ('blob5.a', 'blob5.b', 'blob5.c')]
        eq_(self.archiver.job_blobs, expected)

    @property
    def fake_tag_map(self):
        return {
            'tag1': [
                ('blob1.a', 'blob1.b', 'blob1.c'),
                ('blob2.a', 'blob2.b', 'blob2.c')],
            'tag2': [
                ('blob3.a', 'blob3.b', 'blob3.c'),
                ('blob4.a', 'blob4.b', 'blob4.c'),
                ('blob5.a', 'blob5.b', 'blob5.c')]}

    def test_archive(self):
        incoming_tag = 'incoming:data:chunk:2011-11-13'
        archived_tag = 'processed:data:chunk:2011-11-13'

        # no archived tags before the archive call
        self._setup(tags=[incoming_tag])
        eq_([incoming_tag], self.archiver.ddfs.list(incoming_tag))
        eq_([], self.archiver.ddfs.list(archived_tag))

        # one archived tag after the archive call
        self.archiver.archive()
        eq_([incoming_tag], self.archiver.ddfs.list(incoming_tag))
        eq_([archived_tag], self.archiver.ddfs.list(archived_tag))

        # incoming and archived tags point to the same blobs
        expected_blobs = [
            ('b13.1', 'b13.2', 'b13.3'),
            ('b13.1.a', 'b13.2.a', 'b13.3.a')]
        incoming_blobs = self.archiver.ddfs.blobs(incoming_tag)
        archived_blobs = self.archiver.ddfs.blobs(archived_tag)
        eq_(incoming_blobs, expected_blobs)
        eq_(archived_blobs, expected_blobs)


class TestBuildTagMap(object):

    def _setup(self, archive_mode=True, max_blobs=100, archive_some=False):
        ddfs = DDFS()
        if archive_some:
            blobs = ('b13.1', 'b13.2', 'b13.3')
            ddfs.ddfs['processed:data:chunk:2011-11-13'] = [blobs]
        self.archiver = Archiver(
            ddfs=ddfs,
            archive_prefix='processed',
            archive_mode=archive_mode,
            max_blobs=max_blobs,
            tags=['incoming:data:chunk'])

    def test_tag_map(self):
        self._setup()
        self._assert_full_tag_map()

    def test_partially_archived(self):
        self._setup(archive_some=True)
        self._assert_tag_map_minus_processed()

    def test_archive_mode_off(self):
        self._setup(archive_some=True, archive_mode=False)
        self._assert_full_tag_map()

    def test_max_blobs_zero(self):
        expected = {'incoming:data:chunk:2011-11-14': []}
        self._assert_max_blobs(expected, max_blobs=0)

    def test_max_blobs_some_of_one_tag(self):
        expected = {
            'incoming:data:chunk:2011-11-14': [
                ('b14.1', 'b14.2', 'b14.3'),
                ('b14.1.a', 'b14.2.a', 'b14.3.a')]}
        self._assert_max_blobs(expected, max_blobs=2)

    def test_max_blobs_exactly_one_tag(self):
        expected = {
            'incoming:data:chunk:2011-11-14': [
                ('b14.1', 'b14.2', 'b14.3'),
                ('b14.1.a', 'b14.2.a', 'b14.3.a'),
                ('b14.1.b', 'b14.2.b', 'b14.3.b')]}
        self._assert_max_blobs(expected, max_blobs=3)

    def test_max_blobs_more_than_one_tag(self):
        expected = {
            'incoming:data:chunk:2011-11-13': [
                ('b13.1.a', 'b13.2.a', 'b13.3.a')],
            'incoming:data:chunk:2011-11-14': [
                ('b14.1', 'b14.2', 'b14.3'),
                ('b14.1.a', 'b14.2.a', 'b14.3.a'),
                ('b14.1.b', 'b14.2.b', 'b14.3.b')]}
        self._assert_max_blobs(expected, max_blobs=4)

    def _assert_max_blobs(self, expected, max_blobs):
        self._setup(archive_some=True, max_blobs=max_blobs)
        self._compare_blobs(self.archiver.tag_map, expected)
        eq_(self.archiver.blob_count, max_blobs)

    def _assert_full_tag_map(self):
        expected = {
            'incoming:data:chunk:2011-11-11': [
                ('b11.1', 'b11.2', 'b11.3')],
            'incoming:data:chunk:2011-11-12': [
                ('b12.1', 'b12.2', 'b12.3')],
            'incoming:data:chunk:2011-11-13': [
                ('b13.1', 'b13.2', 'b13.3'),
                ('b13.1.a', 'b13.2.a', 'b13.3.a')],
            'incoming:data:chunk:2011-11-14': [
                ('b14.1', 'b14.2', 'b14.3'),
                ('b14.1.a', 'b14.2.a', 'b14.3.a'),
                ('b14.1.b', 'b14.2.b', 'b14.3.b')]}
        self._compare_blobs(self.archiver.tag_map, expected)
        eq_(self.archiver.blob_count, 7)

    def _assert_tag_map_minus_processed(self):
        expected = {
            'incoming:data:chunk:2011-11-11': [
                ('b11.1', 'b11.2', 'b11.3')],
            'incoming:data:chunk:2011-11-12': [
                ('b12.1', 'b12.2', 'b12.3')],
            'incoming:data:chunk:2011-11-13': [
                ('b13.1.a', 'b13.2.a', 'b13.3.a')],
            'incoming:data:chunk:2011-11-14': [
                ('b14.1', 'b14.2', 'b14.3'),
                ('b14.1.a', 'b14.2.a', 'b14.3.a'),
                ('b14.1.b', 'b14.2.b', 'b14.3.b')]}
        self._compare_blobs(self.archiver.tag_map, expected)
        eq_(self.archiver.blob_count, 6)

    def _compare_blobs(self, tag_map, expected):
        eq_(len(expected), len(tag_map))
        for tag, blob in expected.iteritems():
            ok_(tag in tag_map)
            tag_map_blob = tag_map[tag]
            eq_(len(tag_map_blob), len(blob))
            for replica in blob:
                ok_(replica in tag_map_blob)



import os
import json
import random


def fixture_path(name):
    here = os.path.dirname(__file__)
    return os.path.join(here, '..', 'fixture', name)


class Job(object):

    def __init__(self, name, *args, **kwargs):
        self.name = '%s@%.5f' % (name, random.random())
        self.options = kwargs

    def profile_stats(self):
        raise NotImplemented

    def wait(self):
        pass


class Disco(object):

    def __init__(self, server=None):
        self.server = server
        self.master = server

    def new_job(self, *args, **kwargs):
        return Job(*args, **kwargs)

    def purge(self, job_name):
        pass


class DDFS(object):

    def __init__(self, server=None):
        self.server = server
        with open(fixture_path('ddfs.json')) as f:
            self.ddfs = json.loads(f.read())
        for tag, blobs in self.ddfs.iteritems():
            self.ddfs[tag] = [tuple(x) for x in blobs]

    def list(self, tag):
        return [key for key in self.ddfs.iterkeys() if key.startswith(tag)]

    def blobs(self, tag):
        return self.ddfs.get(tag, [])

    def tag(self, tag, target):
        self.ddfs.setdefault(tag, []).extend(target)



from nose.tools import eq_

from inferno.lib.job_factory import JobFactory
from inferno.lib.rule import InfernoRule
from inferno.lib.settings import InfernoSettings


class TestJobFactory(object):

    def TODO_create(self):
        rule = InfernoRule(
            name='some_rule_name',
            source_tags=['some_tag'],
            some_extra_param='some_extra_value',
            day_range=2,
            day_offset=4)
        settings = InfernoSettings(
            settings_file='some_unknown_settings_file',
            server='some_server',
            day_range=3,
            day_offset=5,
            debug=True,
            disco_debug=True)
        expected_settings = {
            # the passed in overrides
            'day_range': 3,
            'day_offset': 5,
            'debug': True,
            'disco_debug': True,
            'server': 'some_server',
            'settings_file': 'some_unknown_settings_file',
            # the normal defaults
            'rules_directory': '/apps/project/project/rules',
            'extra_python_paths': [],
            'pid_dir': '/var/run/inferno',
            'spawn_delay': 5,
            'max_workers': 8,
            'log_config': '/etc/inferno/log.ini'}
        expected_params = {
            # extra rule kwargs, plus keysets
            'some_extra_param': 'some_extra_value',
            'keysets': {},
            # the passed in overrides
            'day_range': 3,
            'day_offset': 5,
            'debug': True,
            'disco_debug': True,
            'server': 'some_server',
            'settings_file': 'some_unknown_settings_file',
            # the normal defaults
            'rules_directory': '/apps/project/project/rules',
            'extra_python_paths': [],
            'pid_dir': '/var/run/inferno',
            'spawn_delay': 5,
            'max_workers': 8,
            'log_config': '/etc/inferno/log.ini',
            'parts_preprocess': []}
        job = JobFactory.create_job(rule, settings)

        # check disco/ddfs
        eq_(str(job.ddfs), 'DDFS master at disco://some_server')
        eq_(str(job.disco), 'Disco master at disco://some_server')

        # check job.rule
        eq_(job.rule.name, 'some_rule_name')
        eq_(job.rule.source_tags, ['some_tag'])
        eq_(job.rule.day_range, 2)
        eq_(job.rule.day_offset, 4)

        # check job.settings
        eq_(job.settings, expected_settings)

        # check job.params
        eq_(vars(job.params), expected_params)
